val debug: bool ref

(** ATI is parameterized by clauses (from the Kettle module). This is 
 * anything you want to be returned by findChoices, but they must implement 
 * at least the following inteface.  *)
class ['clause] clauseData : 'clause -> object
  method name: string
  method conc: Logic.pred
  method hyps: Logic.pred list
  method maxUnif: int
  method clause: 'clause
end
   
val dclause: unit -> 'clause clauseData -> Pretty.doc

(** The state of the automaton *) 
type 'clause state

(** Set this to the depth up to which ATI should consider the clause heads *)
val atiDepth: int ref

                                        (* Make a state, given a clause to be 
                                         * used as initialization *)
val make: 'clause clauseData -> 'clause state


(** Add a new clause to the automaton *)
val addClause: 'clause state -> 
  clause:'clause clauseData ->
  unit

(** Fetch a clause given its index *)
val getClause: 'clause state -> int -> 'clause clauseData

(** Delete the last clause from the automaton *)
val deleteLastClause: 'clause state -> unit


(** Given a state of the automaton and a goal (with associated substitution), 
 * find a list of clauses that can possibly match the goal *)
val findChoices: 'clause state 
  -> Logic.spred
  -> 'clause clauseData list   
