
(* This is the main prover engine. It implements the inversion proof 
 * procedure to destruct predicates into atoms and then drives the 
 * satisfiability procedures. Finally it constructs the proof *)


val debug : bool ref

val continue_on_error: bool ref
val continue_without_asking: bool ref

val timingFlag: bool ref


val printAssumptions: bool ref
val printWitness: bool ref

(** The maximum number of seconds for proving a goal. 0 means disable *)
val timeout: int ref


type proofgen =                         (* Something to process the proofs *)
    { start: string -> unit;            (** The file name that was passed 
                                         * with -o *)
      startVC: string -> unit;
      assume: (unit -> string) -> Logic.pred -> unit; (** Generate the 
                                                            * name lazily *)
      unassume: unit -> unit;
      pick: string -> unit;
      unpick: unit -> unit;
      conjunct: Logic.pred -> Proof.proof -> unit;
      choose: thechoice:int -> nrchoices:int -> unit;  
      finish: unit -> unit;
      finishVC: unit -> unit}

                                        (* Start a set of proofs *)
val start: string -> Satproc.entries list -> proofgen option -> unit

(** Prove a goal, given a name, a state, a function for reading the elements 
 * of the goal *)
val proveGoal: string -> (unit -> Input.input) -> unit


(** Add a global assumption *)
val globalAssume: Logic.pred -> Errormsg.location -> unit

                                        (* Finish a set of proofs *)
val finish: unit -> unit
                                        (* Raised when proving fails *)
exception Failure

val choose: thechoice:int -> nrchoices:int -> unit


(** Remember the current location **)
val currLoc: Input.loc ref



class type uiType = object
  method name: string
  method options: Options.optionDescr list
  method enterBlock: (Util.symbol * Logic.pred) list -> 
                     (string * string) list -> unit
  method exitBlock: unit

  (** Called when a goal is initiated *)
  method start_goal: Logic.pred -> unit
      
  (** Called when a goal is successfully proved *)
  method goal_success: Logic.pred -> Proof.proof -> unit

  (** Called when a goal fails, with a witness. *)
  method goal_failure: Errormsg.location -> 
                       string -> Logic.pred -> Pretty.doc list -> unit

  (** Called when an assumption is encountered. Generate the name lazily *)
  method assume: (unit -> Util.symbol) -> Logic.pred -> unit

  (** Called when an assumption is retracted *)
  method unassume: unit

  (** Called to start the UI *)  
  method run: unit
end

val theUI: uiType ref

val init: unit -> unit

(** Define new printers that print either infix or prefix *)
val printInfix: bool ref
val d_exp: unit -> Logic.exp -> Pretty.doc
val d_pred: unit -> Logic.pred -> Pretty.doc

val options: Options.optionDescr list


