{
(* This program is free software.  See LICENSE for more information *)
module E = Errormsg
module S = Satproc
module SMT = Smtlib

module I = Input
module L = Logic

open Smtlibparse

(*** Error handling ***)
let lexerror msg =
  E.parse_error msg


}


rule token = parse
  | "true" {TRUE_TOK}
  | "false" {FALSE_TOK}
  | "ite" {ITE_TOK}
  | "not" {NOT_TOK}
  | "implies" {IMPLIES_TOK}
  | "if_then_else" {IF_THEN_ELSE_TOK}
  | "and" {AND_TOK}
  | "or" {OR_TOK}
  | "xor" {XOR_TOK}
  | "iff" {IFF_TOK}
  | "exists" {EXISTS_TOK}
  | "forall" {FORALL_TOK}
  | "let" {LET_TOK}
  | "flet" {FLET_TOK}
  | '"' [^ '"']* '"' (* '"' *) {STRING_TOK(Lexing.lexeme lexbuf)}
  | '{' [^ '{' '}']* '}' {USER_VAL_TOK(Lexing.lexeme lexbuf)}
  | "notes" {NOTES_TOK}
  | "sorts" {SORTS_TOK}
  | "funs" {FUNS_TOK}
  | "preds" {PREDS_TOK}
  | "extensions" {EXTENSIONS_TOK}
  | "definition" {DEFINITION_TOK}
  | "axioms" {AXIOMS_TOK}
  | "logic" {LOGIC_TOK}
  | ':' {COLON_TOK}
  | '(' {LPAREN_TOK}
  | ')' {RPAREN_TOK}
  | "sat" {SAT_TOK}
  | "unsat" {UNSAT_TOK}
  | "unknown" {UNKNOWN_TOK}
  | "assumption" {ASSUMPTION_TOK}
  | "formula" {FORMULA_TOK}
  | "status" {STATUS_TOK}
  | "benchmark" {BENCHMARK_TOK}
  | "extrasorts" {EXTRASORTS_TOK}
  | "extrafuns" {EXTRAFUNS_TOK}
  | "extrapreds" {EXTRAPREDS_TOK}
  | "language" {LANGUAGE_TOK}
  | "=" {EQUALS_TOK}
  | "distinct" {DISTINCT_TOK}
  | ';' [^'\n']* "\n" {token lexbuf}  
  | '$' {DOLLAR_TOK}
  | ['=' '<' '>' '&' '@' '#' '+' '-' '*' '/' '%' '|' '~']+ {AR_SYMB(Lexing.lexeme lexbuf)}
  | ['0'-'9']+ {NUMERAL_TOK(Lexing.lexeme lexbuf)}
  | '?' ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '.' '_' '0'-'9']* {VAR_TOK(Lexing.lexeme lexbuf)}
  | ['a'-'z' 'A'-'Z' '/']['a'-'z' 'A'-'Z' '.' '-' '+' '/' '_' '\'' '0'-'9']* {SYM_TOK(Lexing.lexeme lexbuf)}
  | [' ' '\t' '\n' '\r'] {token lexbuf}
  | eof   {EOF_TOK}
  | _ {token lexbuf} 




{

 
(* Remember here the benchmark *)
let inputs: I.input list ref = ref []
let initParser (fname: string) (lexbuf: Lexing.lexbuf) : unit = 
  Smtlib.initParser ();
  let b = Smtlibparse.benchmark token lexbuf in
  (* Make a list of inputs *)
  inputs := 
    I.StartGoal (SMT.currentBenchmark.SMT.name, E.locUnknown)
    (* Now the assumptions. They are in reverse order already. *)
    :: 
      (List.fold_left 
         (fun acc f -> 
           I.Assume (f, E.locUnknown) :: acc) 
         [ I.Assume (SMT.currentBenchmark.SMT.formula, E.locUnknown);
           I.Prove (L.Bool false, E.locUnknown);
           I.EndGoal E.locUnknown;
           I.Eof 
         ]
         SMT.currentBenchmark.SMT.assumptions);
    
  ()

let parser = 
  { S.language    = "smtlib";
    S.initParser        = initParser;
    S.nextFewInputs = 
    (fun lexbuf -> 
      let i = !inputs in 
      inputs := [];
      i);
    S.parserWitness = 
    (fun f -> 
      List.iter f !SMT.parserWitness);
  } 

}


