%{
//#define SHOW printf("yytext=::%s::\n",yytext)
#define SHOW 
int line_number=0,sum,len,i;
#include "hash.h"

/* We'll keep a hash table of the strings we have seen */
PHASH identifiers;

void initLexer() {
  identifiers = NewStringHash();
}

void finishLexer () {
  FreeHash(identifiers);
}

%}

DIGIT    [0-9]
LETDIG   [0-9a-zA-Z_\-\$\#\@\!\+\=\<\>\,\?\\\~]
CAPITAL  [A-Z]
SMALL    [a-z]
WS       [\t \ ]
PATHNAME [a-zA-Z0-9\.\/\\\-]


%%
"ANN_STARTLF"        { SHOW; }
"\n"                 { SHOW; line_number++; }
"ANN_ENDLF"          { SHOW; }
"\%!Decl"		{ SHOW; return BEGINDECL; }
"\%!EndDecl"		{ SHOW; return ENDDECL; }
"\%!COMM"[^'!''\n']*"\n" { SHOW; line_number++; }
"\%"[^'\n']*"\n"	{ SHOW; line_number++; }
{WS}		{ }
"("		{ SHOW; return LPAREN; }
")"		{ SHOW; return RPAREN; }
"["		{ SHOW; return LBRACKET; }
"]"		{ SHOW; return RBRACKET; }
"{"		{ SHOW; return LBRACE; }
"}"		{ SHOW; return RBRACE; }
"->"		{ SHOW; return ARROW; }
"."		{ SHOW; return DOT; }
":"		{ SHOW; return COLON; }
"+"		{ SHOW; return PLUS; }
"*"		{ SHOW; return TIMES; }
"<"		{ SHOW; return LT; }
">"		{ SHOW; return GT; }
"<="		{ SHOW; return LEQ; }
">="		{ SHOW; return GEQ; }
"="		{ SHOW; return EQ; }
"<>"		{ SHOW; return NEQ; }
"mod"		{ SHOW; return MOD; }
"div"		{ SHOW; return DIV; }
"floor"		{ SHOW; return FLOOR; }
"ceil"		{ SHOW; return CEIL; }
"true"		{ SHOW; return TRUE; }
"false"		{ SHOW; return FALSE; }
"/\\"		{ SHOW; return AND; }
"\\/"		{ SHOW; return OR; }
"=>"		{ SHOW; return IMPL; }
"all"		{ SHOW; return ALL; }
"exp"		{ SHOW; return EXP; }
"pred"		{ SHOW; return PRED; }
"pf"		{ SHOW; return PF; }
"choose"        { SHOW; return CHOOSE; }
{DIGIT}+	{ SHOW; len=strlen(yytext);
                  sum=0;
                  for (i=0;i<len;i++) sum=(sum*10)+yytext[i]-'0';
                  yylval.integer_t=sum; return NUM;
                }
"-"		    { SHOW; return MINUS; }
{SMALL}{LETDIG}*    { 
                      HASH_ENTRY* phe;
                      int found; 
                      SHOW; 
                      phe = HashLookupOrAdd(identifiers,yytext,&found);
                      if (found == 0) { (char*)(phe->key)=strdup(yytext); }
                      yylval.string_t = (char*)(phe->key);
                      return ID; }
{CAPITAL}{LETDIG}*  { SHOW; 
                      yylval.string_t = strdup(yytext);
                      return VAR; 
                    }
.		    { SHOW; 
                      printf("------------------Illegal Character ::%c::\n",
                             yytext[0]); 
                    }
