%{
#include <stdlib.h>
#include <time.h>  
#include <string.h>
#include "main.h"  
#include "input.h"

static struct pred theGoal;
 
void fromList(struct predlist* a, struct pred* b);
void makeAxiom(char* str, struct lf l, struct input* a);
struct pred tmp_pred;
struct input tmp_input;

void yyerror(char* s) {
  fprintf(stderr, "%s\n",s);
}

int yylex(void);

int yywrap(void) {
  return(1);
}
 
%}

%union {
char* string_t;
int integer_t;
struct input input_t;
struct pred pred_t;
struct predlist* predlist_t;
struct exp exp_t;
struct explist* explist_t;
struct atom atom_t;
struct element element_t;

struct lf lf_t;
}

%token <string_t> ID VAR USE COMMENT
%token <integer_t> NUM
%token EXP PRED
%token PF
%token LPAREN RPAREN LBRACKET RBRACKET LBRACE RBRACE
%token DOT COLON ARROW
%token PLUS MINUS TIMES
%token MOD DIV FLOOR CEIL
%token LT GT LEQ GEQ EQ NEQ
%token TRUE FALSE AND OR IMPL ALL EXISTS CHOOSE
%token BEGINDECL ENDDECL
%token EOF STARTLF

%start main1

%type <input_t> main
%type <pred_t> pred
%type <predlist_t> predlist
%type <exp_t> exp
%type <explist_t> exps
%type <string_t> bind
 
%type <lf_t> lfatom lfterm
%type <string_t> lfconstant


%type <atom_t> conjunct
%type <element_t> element
%%


main1: VAR COLON PF pred DOT         { theGoal = $4; }
;

final:
    main    { }
  | element { }
;

main:  
    VAR COLON PF                     { $$.tag=GOAL_INPUT;
                                       $$.u.str=strdup($1); }
  | LPAREN CHOOSE NUM NUM RPAREN     { $$.tag=TOPCHOOSE_INPUT;
                                       $$.u.s1.i1=$3; $$.u.s1.i2=$4; }
  | lfconstant COLON lfterm DOT      { makeAxiom($1,$3,&tmp_input);
                                       $$=tmp_input; }
  | PF COLON lfterm DOT              { $$.tag=DECL_INPUT; $$.u.str = "pf"; }
  | USE				     { $$.tag=IMPORT_INPUT;
                                       $$.u.str= strdup($1); }
  | EOF                              { $$.tag=EOF_INPUT; }
;
   
element: 
    conjunct                         { $$.tag=CONJUNCT_ELEMENT; $$.u.a=$1; }
  | LPAREN CHOOSE NUM NUM RPAREN     { $$.tag=CHOOSE_ELEMENT; $$.u.s1.i1=$3; $$.u.s1.i2=$4; }
  | LPAREN IMPL conjunct             { $$.tag=ASSUME_ELEMENT; $$.u.a=$3; }
  | RPAREN                           { $$.tag=RETRACT_ELEMENT; }
  | COMMENT                          { $$.tag=COMMENT_ELEMENT;
                                       $$.u.str = strdup($1); }
  | error RPAREN                     { $$.tag=RETRACT_ELEMENT; }
  | error DOT                        { $$.tag=ENDPROOF_ELEMENT; }
  | DOT                              { $$.tag=ENDPROOF_ELEMENT; }
;

conjunct:
    FALSE				{ $$.tag=FALSE_ATOM; }
  | TRUE                                { $$.tag=TRUE_ATOM; }
  | LPAREN EQ  exp exp	RPAREN 		{ $$.tag=EQ_ATOM; $$.u.s1.e1=$3; $$.u.s1.e2=$4; }
  | LPAREN NEQ exp exp	RPAREN 		{ $$.tag=NEQ_ATOM; $$.u.s1.e1=$3; $$.u.s1.e2=$4; }
  | LPAREN GEQ exp exp	RPAREN 		{ $$.tag=GEQ_ATOM; $$.u.s1.e1=$3; $$.u.s1.e2=$4; }
  | LPAREN GT  exp exp	RPAREN 		{ $$.tag=GT_ATOM; $$.u.s1.e1=$3; $$.u.s1.e2=$4; }
  | LPAREN LEQ exp exp	RPAREN 		{ $$.tag=LEQ_ATOM; $$.u.s1.e1=$3; $$.u.s1.e2=$4; }
  | LPAREN LT  exp exp	RPAREN 		{ $$.tag=LT_ATOM; $$.u.s1.e1=$3; $$.u.s1.e2=$4; }
  | LPAREN ID exps RPAREN		{ $$.tag=BFUN_ATOM;
                                          $$.u.s2.str = $2;
                                          $$.u.s2.elist=$3; }
;
  

/**** Parsing of LF terms in normal form. Applications must be parenthesized.
 ****/
lfconstant: 
    ID     { $$ = $1; }  
  | AND    { $$ = "/\\"; }
  | OR     { $$ = "\\/"; }
  | IMPL   { $$ = "=>"; }
  | ALL    { $$ = "all"; }
  | EQ     { $$ = "="; }
  | NEQ    { $$ = "<>"; }
  | GEQ    { $$ = ">="; }
  | GT     { $$ = ">"; }
  | LEQ    { $$ = "<="; }
  | LT     { $$ = "<"; }
  | FLOOR  { $$ = "floor"; }
  | MOD    { $$ = "mod"; }
  | CEIL   { $$ = "ceil"; }
  | DIV    { $$ = "div"; }
  | EXP    { $$ = "exp"; }
  | PRED   { $$ = "pred"; }
;

lfatom: 
    lfconstant                          { $$.tag=LCONST_LF;
                                          $$.u.str = strdup($1); }
  | LPAREN lfconstant lftermlist RPAREN { $$.tag=LAPP_LF; printf("Unimplemented\n"); }
  | VAR                           { $$.tag=LVAR_LF; printf("Unimplemented\n"); }
  | LPAREN VAR lftermlist RPAREN  { $$.tag=LAPP_LF; printf("Unimplemented\n"); }

  | PF pred                       { $$.tag=LPF_LF; $$.u.p=$2; }
  | LPAREN lfatom RPAREN          { $$=$2; }
;

lfterm:
    lfbindings            { printf("Unimplemented\n"); exit(0); }
;

lftermlist:
  | lfterm            { printf("Unimplemented\n"); exit(0); }
  | lfterm lftermlist { printf("Unimplemented\n"); exit(0); }
; 

lfbindings: 
    lfatom            { printf("Unimplemented\n"); exit(0); }
  | lfbind lfbindings { printf("Unimplemented\n"); exit(0); }
;
 
lfbind:   
    LBRACE VAR COLON lfterm RBRACE          { printf("Unimplemented\n"); exit(0); }
  | LBRACKET VAR COLON lfterm RBRACKET      { printf("Unimplemented\n"); exit(0); }
  | LPAREN lfterm RPAREN ARROW              { printf("Unimplemented\n"); exit(0); }
  | lfatom ARROW                            { printf("Unimplemented\n"); exit(0); }
;

pred:
    TRUE { $$.tag=TRUE_PRED; }
  | FALSE { $$.tag=FALSE_PRED; }
  | VAR { $$.tag=ATOM_PRED; $$.u.a.tag=BFUN_ATOM; $$.u.a.u.s2.str="pvar"; $$.u.a.u.s2.elist=NULL; }
  | LPAREN VAR VAR RPAREN { $$.tag=ATOM_PRED; $$.u.a.tag=BFUN_ATOM; $$.u.a.u.s2.str = "pvarapp"; $$.u.a.u.s2.elist=NULL; }
  | LPAREN ALL VAR RPAREN { $$.tag=ATOM_PRED; $$.u.a.tag=BFUN_ATOM; $$.u.a.u.s2.str = "pvarall"; $$.u.a.u.s2.elist=NULL; }
  | LPAREN ALL LPAREN bind predlist RPAREN { printf("Unimplemented\n"); exit(0); }
  | LPAREN EXISTS LPAREN bind predlist RPAREN RPAREN { printf("Unimplemented\n"); exit(0); }
  | LPAREN AND predlist	RPAREN 	{
      fromList($3,&tmp_pred);
      $$=tmp_pred; 
    }
  | LPAREN OR pred pred	RPAREN	{ 
      $$.tag=OR_PRED;
      $$.u.s2.p1=(struct pred*)malloc(sizeof(struct pred));
      *($$.u.s2.p1)=$3; 
      $$.u.s2.p2=(struct pred*)malloc(sizeof(struct pred));
      *($$.u.s2.p2)=$4; 
    }
  | LPAREN IMPL pred predlist RPAREN  { 
      $$.tag=IMPL_PRED;
      $$.u.s2.p1=(struct pred*)malloc(sizeof(struct pred));
      *($$.u.s2.p1)=$3;
      fromList($4,&tmp_pred);
      $$.u.s2.p2=(struct pred*)malloc(sizeof(struct pred));
      *($$.u.s2.p2)=tmp_pred;
    }
  | LPAREN EQ  exp exp	RPAREN { 
      $$.tag=ATOM_PRED;
      $$.u.a.tag=EQ_ATOM;
      $$.u.a.u.s1.e1=$3;
      $$.u.a.u.s1.e2=$4;
    }
  | LPAREN NEQ exp exp	RPAREN { 
      $$.tag=ATOM_PRED;
      $$.u.a.tag=NEQ_ATOM;
      $$.u.a.u.s1.e1=$3;
      $$.u.a.u.s1.e2=$4;
    }
  | LPAREN GEQ exp exp	RPAREN {
      $$.tag=ATOM_PRED;
      $$.u.a.tag=GEQ_ATOM;
      $$.u.a.u.s1.e1=$3;
      $$.u.a.u.s1.e2=$4;
    }
  | LPAREN GT  exp exp	RPAREN { 
      $$.tag=ATOM_PRED;
      $$.u.a.tag=GT_ATOM;
      $$.u.a.u.s1.e1=$3;
      $$.u.a.u.s1.e2=$4;
    }
  | LPAREN LEQ exp exp	RPAREN { 
      $$.tag=ATOM_PRED;
      $$.u.a.tag=EQ_ATOM;
      $$.u.a.u.s1.e1=$3;
      $$.u.a.u.s1.e2=$4;
    }
  | LPAREN LT  exp exp	RPAREN { 
      $$.tag=ATOM_PRED;
      $$.u.a.tag=LT_ATOM;
      $$.u.a.u.s1.e1=$3;
      $$.u.a.u.s1.e2=$4;
    }
  | LPAREN ID exps RPAREN {
      $$.tag=ATOM_PRED;
      $$.u.a.tag=BFUN_ATOM;
      $$.u.a.u.s2.str = $2;
      $$.u.a.u.s2.elist=$3;
    }
;

predlist:
    { $$=NULL; }
| pred predlist { 
     $$=(struct predlist*)malloc(sizeof(struct predlist));
     $$->p = $1; $$->next=$2;
}
;

exp: 
    VAR { printf("Unimplemented\n"); exit(0); }
  | NUM { $$.tag=INTEGER_EXP; $$.u.integer = $1; }
  | PLUS exp exp { 
      $$.tag=PLUS_EXP; 
      $$.u.exps.e1 = (struct exp*) malloc(sizeof(struct exp)); 
      *($$.u.exps.e1)=$2;
      $$.u.exps.e2 = (struct exp*) malloc(sizeof(struct exp)); 
      *($$.u.exps.e2)=$3;
    } 
  | MINUS exp exp { 
      $$.tag=MINUS_EXP; 
      $$.u.exps.e1 = (struct exp*) malloc(sizeof(struct exp)); 
      *($$.u.exps.e1)=$2;
      $$.u.exps.e2 = (struct exp*) malloc(sizeof(struct exp)); 
      *($$.u.exps.e2)=$3;
    } 
  | TIMES exp exp { 
      $$.tag=TIMES_EXP; 
      $$.u.exps.e1 = (struct exp*) malloc(sizeof(struct exp)); 
      *($$.u.exps.e1)=$2;
      $$.u.exps.e2 = (struct exp*) malloc(sizeof(struct exp)); 
      *($$.u.exps.e2)=$3;
    } 
  | ID { 
      $$.tag=VAR_EXP;
      $$.u.var = $1; 
    }
  | LPAREN ID exps RPAREN {
      $$.tag=FUN_EXP;
      $$.u.var = $2; 
      $$.u.func.elist = $3;
    } 
  | DIV exp NUM { 
      $$.tag = DIV_EXP;
      $$.u.s.i = $3;
      $$.u.s.e = (struct exp*) malloc(sizeof(struct exp));
      *($$.u.s.e)=$2;
    }
  | MOD exp NUM {
      $$.tag = MOD_EXP;
      $$.u.s.i = $3;
      $$.u.s.e = (struct exp*) malloc(sizeof(struct exp));
      *($$.u.s.e)=$2;
    }
  | FLOOR exp NUM { 
      $$.tag = FLOOR_EXP;
      $$.u.s.i = $3;
      $$.u.s.e = (struct exp*) malloc(sizeof(struct exp));
      *($$.u.s.e)=$2;
    }
  | CEIL exp NUM { 
      $$.tag = CEIL_EXP;
      $$.u.s.i = $3;
      $$.u.s.e = (struct exp*) malloc(sizeof(struct exp));
      *($$.u.s.e)=$2;
    }
  | LPAREN exp RPAREN { $$ = $2; }
;

exps:
    exp { 
      $$=(struct explist*) malloc(sizeof(struct explist));
      $$->e=$1; $$->next=NULL;
   }   
  | exp exps {
      $$=(struct explist*) malloc(sizeof(struct explist));
      $$->e=$1; $$->next=$2;
   }
;

bind:
    LBRACKET VAR COLON EXP RBRACKET { }
;
%%
#include "fast_lexer.c"

struct pred* parseGoal(char *fname) {
  int res;
  if ((yyin=fopen(fname,"r"))==NULL) {
    _ERROR1("Cannot open file %s\n", fname);
  }
  fprintf(stderr, "Parsing file %s\n", fname);
  initLexer();
  res = yyparse();
  finishLexer();
  fclose(yyin);
  if(res == 0) {
    return &theGoal;
  } else
    return NULL;
}




void fromList(struct predlist* a, struct pred* b) 
{ 
   struct pred tmp;
   if (a == NULL) { b->tag=TRUE_PRED; } 
   else if (a->next==NULL) { *b=a->p; }
   else while (a != NULL) {
      b->tag=AND_PRED;
      b->u.s2.p1=(struct pred*)malloc(sizeof(struct pred)); 
      *(b->u.s2.p1)=a->p;
      fromList(a->next,&tmp);
      b->u.s2.p2=(struct pred*)malloc(sizeof(struct pred));
      *(b->u.s2.p2)=tmp;
   }
}

void makeAxiom(char* str, struct lf l, struct input* a)
{
   if ((l.tag==LPF_LF)&&(l.u.p.tag==ATOM_PRED)) {
       a->tag=AXIOM_INPUT; 
       a->u.a.str = strdup(str); 
       a->u.a.vlist=NULL; 
       a->u.a.alist=NULL; 
       a->u.a.a=l.u.p.u.a;
   };
}
