#include "input.h"

void print_exp (struct exp* e)
{
  struct explist* el;
  switch (e->tag) {
  case VAR_EXP: printf("%s",e->u.var); break;
  case INTEGER_EXP: printf("%d",e->u.integer); break;
  case TIMES_EXP: print_exp(e->u.exps.e1); printf(" * "); print_exp(e->u.exps.e2); break;
  case PLUS_EXP: print_exp(e->u.exps.e1); printf(" + "); print_exp(e->u.exps.e2); break;
  case MINUS_EXP: print_exp(e->u.exps.e1); printf(" - "); print_exp(e->u.exps.e2); break;
  case FUN_EXP:
    el=e->u.func.elist; 
    printf("%s(",e->u.func.var);
    if (el != NULL) {
      print_exp(&(e->u.func.elist->e)); el = el->next;
      while (el != NULL) {
	printf(", "); print_exp(&(el->e));
	el=el->next;
      }
    }
    printf(")");
    break;
  default: printf("Unimplemented\n"); exit(1);
  }
}

void print_atom (struct atom a)
{
  switch (a.tag) {
  case FALSE_ATOM: printf("false"); break;
  case TRUE_ATOM: printf("true"); break;
  case EQ_ATOM: print_exp(&(a.u.s1.e1)); printf(" = "); print_exp(&(a.u.s1.e2)); break;
  case NEQ_ATOM: print_exp(&(a.u.s1.e1)); printf(" != "); print_exp(&(a.u.s1.e2)); break;
  case GEQ_ATOM: print_exp(&(a.u.s1.e1)); printf(" >= "); print_exp(&(a.u.s1.e2)); break;
  case GT_ATOM: print_exp(&(a.u.s1.e1)); printf(" > "); print_exp(&(a.u.s1.e2)); break;
  case LEQ_ATOM: print_exp(&(a.u.s1.e1)); printf(" <= "); print_exp(&(a.u.s1.e2)); break;
  case LT_ATOM: print_exp(&(a.u.s1.e1)); printf(" < "); print_exp(&(a.u.s1.e2)); break;
  case BFUN_ATOM: break;
  case NOTBFUN_ATOM: break;
  }
}

void print_pred (struct pred p)
{
  switch (p.tag) {
  case TRUE_PRED: printf("true pred"); break;
  case FALSE_PRED: printf("false pred"); break;
  case ALL_PRED: printf("Unimplemented"); break;
  case EXIST_PRED: printf("Unimplemented"); break;
  case AND_PRED: print_pred(*(p.u.s2.p1)); printf(" and_pred "); print_pred(*(p.u.s2.p2)); break;
  case OR_PRED: print_pred(*(p.u.s2.p1)); printf(" or_pred "); print_pred(*(p.u.s2.p2)); break;
  case IMPL_PRED: print_pred(*(p.u.s2.p1)); printf(" impl_pred "); print_pred(*(p.u.s2.p2)); break;
  case ATOM_PRED: printf("atom pred: "); print_atom(p.u.a); break;
  }
}

void print_predlist (struct predlist* pl)
{
  while (pl != NULL) {
    print_pred(pl->p); printf(" ");
    pl=pl->next;
  }
}
      
