#!/bin/perl
use strict;

#
### Write here the command to use, with __FILE__ to be 
### substituted by the filename
#

#my $comm = "make EXTRAARGS='-timeout 4' prove/__FILE__";
my $comm = "(make EXTRAARGS='-timeout 1 -oldbug' prove/__FILE__; make EXTRAARGS='-timeout 1' prove/__FILE__)";


###
### Copy below lines that must be matched. Each line below is considered as a
### substring that must appear in the output. 
###
my $expectedLines = '
Bug: Got contradiction
Kettle success
';
#Error: Proof checking failed:
#Interrupted: Timeout


###
### Copy below lines that must not appear (none of these lines may appear in a
### succesful test)
###
my $forbiddenLines = '

';

###############################################################
####### SHOULD NOT NEED TO EDIT BELOW #########################
###############################################################
my $basefile = $ARGV[0];
my $file = "$basefile.cvc";
if(! -f $file) { 
    die "Argument must be a the basename of a CVC file"; 
}

# We'll use two copies of the file xxx_tmp.cvc and xxx_min.cvc
#
my $tmpfile = $basefile. "_tmp";
my $minfile = $basefile . "_min";
my $outfile = $basefile . ".minout";


sub command {
    my ($fname) = @_;
    my $res = $comm;
    $res =~ s|__FILE__|$fname|g;
    return $res;
}

sub pause { 
    my ($msg) = @_;
    # print "$msg. Press a key"; <STDIN>;
}

sub printListAsRanges {
    my (@lst) = @_;
#    print "(", join(',', @lst), ") ";
    my $start = shift @lst;
    if(!defined $start) { return; }
    print $start;
    my $last = $start;
    while(1) {
        my $next = shift @lst;
        if(! defined $next) { 
            if($last > $start) {
                print "-$last";
            } 
            return; 
        }
        if($next > $last + 1) {
            if($last != $start) {
                print "-$last";
            }
            print ",$next";
            $start = $next;
        }
        $last = $next;
    }
}

# These are all the regexps we are looking for
my @expectedregexps;

# Make a bunch of regexps from the 
foreach my $l (split /\n/,$expectedLines) {
   if($l eq "") { next; }
   # Drop the surrounding white space
   $l =~ s/^\s*//;
   $l =~ s/\s*$//;
   # print "Looking for:$l.\n";
   push @expectedregexps, quotemeta $l;
}

if (@expectedregexps == 0) {
   print "SUCCESS. No patterns to look for\n"; exit 0;
}

my @forbiddenregexps;

# Make a bunch of regexps from the forbiddenLines
foreach my $l (split /\n/,$forbiddenLines) {
   if($l eq "") { next; }
   # Drop the surrounding white space
   $l =~ s/^\s*//;
   $l =~ s/\s*$//;
   # print "Looking for:$l.\n";
   push @forbiddenregexps, quotemeta $l;
}

sub checkOutput { 
    my ($command, $outfile) = @_;
    my $debug = 1;

    my @regexps = @expectedregexps;
    open(OUTPUT, "$command 2>&1 |") || die "Cannot run $command";
    open(COPY, ">$outfile") || die "Cannot open $outfile";
    my $count = -1;
    my $res = "";
    while(<OUTPUT>) {
        #if($debug) { print $_; }
        $count = $count + 1;
        print COPY $_;
        foreach my $r (@regexps) {
            if($_ =~ m|$r|) {
                if($debug) { "Found $r in line $count\n"; }
                # Remove $r from @regexps
                @regexps = grep { $_ ne $r } @regexps;
                if (@regexps == 0) {
                    goto Done;
                }
            }
        }
        foreach my $f (@forbiddenregexps) {
            if($_ =~ m|$f|) {
                if($debug) { "Found forbidden $f in line $count\n"; }
                $res = $f;
                goto Done;
            }
        }
    }
    
    $res = join('|', @regexps);
  Done: 
    close(COPY);
    close(OUTPUT);
    return $res;
}

# Make a copy of the file
sub copy {
    my ($src, $dest) = @_;
    open(IN, "<$src.cvc") || die "Cannot read from $src.cvc";
    open(OUT, ">$dest.cvc") || die "Cannot write to $dest.cvc";

    while(<IN>) {
        print OUT $_;
    }
    close(OUT);
    close(IN);

}

# Make a copy of the file and clean comments, empty blocks; Count asserts
# Move RETRACT to be each on one line
sub copyCleanCountAsserts {
    my ($src, $dest) = @_;
    open(IN, "<$src.cvc") || die "Cannot read from $src.cvc";
    open(OUT, ">$dest.cvc") || die "Cannot write to $dest.cvc";

    my $countasserts = 0;

    # Keep here the last blocks
    my @blocks = ();

    while(<IN>) {
        # Move RETRACT to be one on a line
        while($_ =~ m|^\s*RETRACT;(.*)$|) {
            $_ = $1;
            print OUT "RETRACT;\n";
        }
        if($_ =~ m|^\s*$|) { # empty line
            next;
        }
        if($_ =~ m|^\s*%|) { # comment line
            next;
        }
        if($_ =~ m|^\s*(BLOCK.+)$|) {
            push @blocks, $1;
            next;
        }
        if($_ =~ m|^\s*ENDBLOCK$|) {
            if(@blocks > 0) {
                pop @blocks; # remove one block
            } else {
                print OUT $_;
                print OUT "\n";
            }
            next;
        }
        # An interesting line
        foreach my $b (@blocks) {
            print OUT "$b\n";
        }
        @blocks = ();
        
        if($_ =~ m|^\s*ASSERT|) {
            $countasserts ++;
        } 
        print OUT $_;
    }
    close(OUT);
    close(IN);

    &pause("copyCleanCountAsserts($src, $dest) = $countasserts");
    return $countasserts;
}


# Make a copy of the file and enable a list of ASSERT indices, starting at 0,
# and sorted. Enable only the corresponding RETRACTs
sub copyEnableAsserts {
    my ($src, $dest, @toenable) = @_;
    open(IN, "<$src.cvc") || die "Cannot read from $src.cvc";
    open(OUT, ">$dest.cvc") || die "Cannot write to $dest.cvc";

    my $countasserts = -1;

    # Make a copy of toenable for the RETRACT
    my @toenableretract = @toenable;

    # Keep a stack of unmatched ASSERTs
    my @unmatchedASSERT = ();

    while(<IN>) {
        # Look for a possibly commented assert
        if($_ =~ m|^\s*(%)?(ASSERT.*)$|) {
            my $theassert = $2;
            $countasserts ++;

            push @unmatchedASSERT, $countasserts;

            # See if we must enable it
            my $disable = 1;
            my $next = shift @toenable;
            if(defined $next) {  
                if($next == $countasserts) { 
                    $disable = 0;
                } else { 
                    unshift @toenable, $next;
                }
            }
            if($disable) { print OUT "%"; }
            print OUT "$theassert\n";
            next;
        } 
        if($_ =~ m|^\s*(%)?RETRACT;(.*)$|) {
            if($2 ne "") { die "Did not expect stuff after RETRACT"; }
            # Find to which ASSERT this one corresponds
            my $thisassert = pop @unmatchedASSERT;
            # print "dealing with RETRACT for asssert no $thisassert\n";

            if(! defined $thisassert) { die "unmatched RETRACT"; }
            # This RETRACT corresponds to $countasserts
            my $enable = 0;
            foreach my $r (@toenableretract) {
                if($r == $thisassert) {
                    $enable = 1; last;
                } elsif($r > $thisassert) { 
                    last;
                }
            }
            if($enable) { 
                print OUT "RETRACT;\n";
            } else {
                print OUT "%RETRACT;\n";
            }
            next;
        }
        print OUT $_;
    }
    close(OUT);
    close(IN);
    &pause("copyEnableAsserts($src, $dest, ". join(',', @toenable));
}


# merge two sorted lists
sub mergeLists { 
    my ($pl1, $pl2) = @_;
    my @l1 = @{$pl1};
    my @l2 = @{$pl2};

    my @res;

    my $h1 = shift @l1;
    my $h2 = shift @l2; # The current heads
    while(1) { 
        if(! defined $h1 && ! defined $h2) { last; }
        if(defined $h1 && (! defined $h2 || $h1 < $h2)) {
            push @res, $h1; $h1 = shift @l1; 
            next;
        }
        if(defined $h2 && (! defined $h1 || $h2 < $h1)) {
            push @res, $h2; $h2 = shift @l2; 
            next;
        }
        die "Should not get here";
    }
    return @res;
}


# Try a test
sub tryTest {
    my(@enable) = @_;

    print "  Testing with asserts: "; &printListAsRanges(@enable); print "\n";

    &copyEnableAsserts($minfile, $tmpfile, @enable);
    my $res = &checkOutput(&command($tmpfile), $outfile);
    if($res eq "") { 
        print "     success\n";
        &copy($tmpfile, $minfile); # Try to keep a small file around
        return 1;
    } else {
        print "     failed\n";
        return 0;
    }
}

my $nrasserts; # Total number of asserts

# There are a number of candidates for removal, and a few tests that are
# forced to exist. Must compute the minimal set of candidates that must be
# present for the test to behave the same as with all candidates and forced
# tests. The two lists are sorted. You know that if we have only the pforced
# (drop all candidates) then the test fails. 
sub deltaDebug { 
    my ($pcandidates, $pforced) = @_;

    my @candidates = @{$pcandidates};
    print "Try to minimize "; 
    &printListAsRanges(@candidates);
    print "\n  while forcing ";
    &printListAsRanges(@{$pforced});
    print "\n";

    my $nrcandidates = @candidates;
    if($nrcandidates == 0) {
        return ();
    }

    # If the nrcandidates was 1 then firsthalf will be empty. We know that
    # with the single candidate we succeed, and with just the forced we fail.
    # So we must keep the candidate
    if($nrcandidates == 1) { 
        return @candidates;
    }

    # Split the candidates roughly in half
    my @firsthalf;
    for (my $i=0; $i < $nrcandidates / 2; $i ++) {
        push @firsthalf, (shift @candidates);
    }
    my @secondhalf = @candidates;
    if(@firsthalf == 0 || @secondhalf == 0) { die "Why?"; }

    # Test the first half + forced ones
    my @try_without_secondhalf = &mergeLists(\@firsthalf, $pforced);
    # Do this test
    if(&tryTest(@try_without_secondhalf)) {
        # Test succeeds; We can drop all of the second half and maybe some of
        # the first half as well
        &deltaDebug(\@firsthalf, $pforced);
    } else { 
        # Test the first half + forced one
        my @try_without_firsthalf = &mergeLists(\@secondhalf, $pforced);
        if(&tryTest(@try_without_firsthalf)) {
            # Test succeeds; We can drop all of first half, and maybe some of
            # the second half as well
            &deltaDebug(\@secondhalf, $pforced);
        } else {
            # Both tests have failed. 
            my @keep_from_firsthalf = &deltaDebug(\@firsthalf,
                                                  \@try_without_firsthalf);

            my @try_without_secondhalf_new =
                &mergeLists(\@keep_from_firsthalf, 
                            $pforced);
            my @keep_from_secondhalf = &deltaDebug(\@secondhalf,
                                                   \@try_without_secondhalf_new);

            my @res = &mergeLists(\@keep_from_firsthalf, 
                                  \@keep_from_secondhalf);
            return @res;
        }
    }

}


    # run the file once to make sure we get the expected output
print "Run once to make sure that we succeed with all asserts\n";
my $res0 = &checkOutput(&command($basefile), $outfile);
if($res0 ne "") { 
    die "Initial file does not match the following patterns\n$res0";
}

# Put a clean copy of the file in the min file
$nrasserts = &copyCleanCountAsserts($basefile, $minfile); # No changes

   # run without any asserts, to make sure we fail
print "Run once with no ASSERT to make sure that we fail\n";
if(&tryTest()) {
    die "This test succeeds even if I remove all patterns\n";
}

# Make a list of appropriate number of asserts
my @allasserts;
for(my $i = 0; $i < $nrasserts; $i ++) {
    push @allasserts, $i;
}

my @keep = &deltaDebug(\@allasserts, []);

# Enable the final batch
&copyEnableAsserts($minfile, $tmpfile, @keep);

# And clean the file
&copyCleanCountAsserts($tmpfile, $minfile);

print "We are done minimizing. Result is in $minfile\n";
print "Out of $nrasserts we kept ", int(@keep), ": ";
&printListAsRanges(@keep);
print "\n";

unlink $tmpfile;
unlink $outfile;
exit 0;

