  /* c2v_ext.lex
   * extension to vml.lex to parse C2V
   */


/* state for consuming embedded C code */  
%x LITCODE

%%

  /* c2v keywords */
"emit"                           return tok(C2VTOK_EMIT);
"for_each_field"                 return tok(C2VTOK_FOR_EACH_FIELD);
"of"                             return tok(C2VTOK_OF);
"return"                         return tok(C2VTOK_RETURN);


  /* c2v embed-begin keywords */
"matchExpr" {
  recentMatchKeyword = true;
  return tok(C2VTOK_MATCH_EXPR);
}

"matchStmt" {
  recentMatchKeyword = true;
  return tok(C2VTOK_MATCH_STMT);
}

"matchType" {
  recentMatchKeyword = true;
  return tok(C2VTOK_MATCH_TYPE);
}

  /* embedded sequences */
"{" {
  if (!recentMatchKeyword) {
    return tok(C2VTOK_LBRACE);
  }
  else {
    recentMatchKeyword = false;
    BEGIN(LITCODE);            
    
    // set up embedded reader
    embedded->reset();
    embedStart = '{';
    embedFinish = '}';

    tok(TOK_LBRACE);
    embedStartLoc = nextLoc;
    return TOK_LBRACE;
  }
}

  /* --------- embedded literal code --------- */
  /* copied/modified from elkhound/gramlex.lex */
<LITCODE>{
  [^\]\})\n]+ {
    updLoc();
    embedded->handle(yytext, yyleng, embedFinish);
  }

  "\n" {
    whitespace();
    embedded->handle(yytext, yyleng, embedFinish);
  }

  ("]"|"}"|")") {
    if (embedded->zeroNesting()) {
      // done
      BEGIN(INITIAL);

      // check for balanced delimiter
      if (embedFinish != yytext[0]) {
        err("unbalanced literal code delimiter");
      }

      // put the delimiter back so we will see it next time
      yyless(0);     // put back all but 0, i.e., put it all back

      // make text available to caller
      sval = (SemanticValue)strtable(embedded->text);

      // reset some state
      embedStart = embedFinish = 0;

      loc = embedStartLoc;
      return C2VTOK_EMBEDDED_CODE;
    }
    else {
      // delimeter paired within the embedded code, mostly ignore it
      updLoc();
      embedded->handle(yytext, yyleng, embedFinish);
    }
  }

  <<EOF>> {
    err(stringc << "hit end of file while looking for final `"
                << embedFinish << "'");
    yyterminate();
  }
}


  /* C2V variable */
"$"[a-z]{ALNUM}*         return svalTok(C2VTOK_VARNAME);

  /* C/VML pattern variables */
"$I"{ALNUM}*             return svalTok(C2VTOK_PATV_IDENTIFIER);

  /* C pattern variables */
"$S"{ALNUM}*             return svalTok(C2VTOK_CPATV_STATEMENT);
"$E"{ALNUM}*             return svalTok(C2VTOK_CPATV_EXPRESSION);
"$T"{ALNUM}*             return svalTok(C2VTOK_CPATV_TYPE);
"$N"{ALNUM}*             return svalTok(C2VTOK_CPATV_VARIABLE);
"$F"{ALNUM}*             return svalTok(C2VTOK_CPATV_FIELD);

  /* VML pattern variables */
"$VN"{ALNUM}*            return svalTok(C2VTOK_VPATV_VARIABLE);
"$VE"{ALNUM}*            return svalTok(C2VTOK_VPATV_EXPRESSION);
"$VP"{ALNUM}*            return svalTok(C2VTOK_VPATV_PREDICATE);
"$VS"{ALNUM}*            return svalTok(C2VTOK_VPATV_STATEMENT);
"$VF"{ALNUM}*            return svalTok(C2VTOK_VPATV_FIELD);

  /* illegal pattern variable */
"$"{ALNUM}* {
  err(stringf("illegal pattern variable name: %.*s", yyleng, yytext).c_str());
  return svalTok(C2VTOK_PATV_IDENTIFIER);     // error recovery
}


  /* EOF: c2v_ext.lex */
