// c2v_parse_c.h
// parse the embedded C in a C2V AST

#ifndef C2V_PARSE_C_H
#define C2V_PARSE_C_H

class C2VFile;              // c2v.ast
class StringTable;          // strtable.h
class CCLang;               // cc_lang.h

// parse all the embedded C in 'file', throwing 'xbase' if
// there are any problems
void c2v_parse_c(C2VFile *file, StringTable &strTable, CCLang &lang);


// the following declarations are for c2v_parse_c.cc only
#ifdef C2V_PARSE_C_IMPL

#include "c2v.ast.gen.h"    // C2VVisitor, C2V AST, etc.
#include "cc_ast.h"         // ASTVisitor
#include "owner.h"          // Owner
#include "glr.h"            // GLR
#include "cc.gr.gen.h"      // CCParse

// walk the C2V AST in search of C fragments, and parse them
class C2VParseC : public C2VVisitor {
public:      // data
  StringTable &strTable;
  CCLang &lang;

  CCParse parseContext;
  Owner<ParseTables> tables;
  GLR glr;

  int errors;

public:      // funcs
  C2VParseC(StringTable &strTable, CCLang &lang);
  virtual ~C2VParseC();

  virtual bool visitC_Statement(C_Statement *obj);
  virtual bool visitC_Expression(C_Expression *obj);
  virtual bool visitC_TypeId(C_TypeId *obj);

  SemanticValue parseFragment(int preToken, SourceLoc loc, StringRef text);
};

// walk the C pattern fragment AST in search of ambiguous nodes,
// and apply disambiguation criteria specifically intended for the
// pattern language
class PatternDisambig : public ASTMVisitor {
public:      // funcs
  virtual bool visitExpression(Expression *&obj);
};

#endif // C2V_PARSE_C_IMPL

#endif // C2V_PARSE_C_H
