#include "selupd.model"     // sel/upd
#include "state.model"      // metaVar

// typing axioms should not refer to objct
verifier_option((no_objct));


typedef enum { NOT_SEEN, /* Means that we have not yet reached this node */
               DONELEFT, /* We have gotten to this node, have marked it,
                          * have saved in its left field the address of the
                          * pointer (c==0) and we are working in its old left
                          * field */
               DONERIGHT, /* We have gotten to this node, have marked it,
                           * have done its left field, have saved in its
                           * right field the address of the parent (c==1),
                           * and are working in its old right field. The
                           * left field is as before. */
               DONE       /* We have gotten to this node, have marked it,
                           * and have processed both its left and its right
                           * fields. The fields are as they were before */
} states;

typedef struct node {
  int m, c;
  struct node *l, *r;
  struct node *ol, *or; // The original fields
  states state;
} *node;


int global_inv() {
  return vc_expression(all_of {
    #if 0      // not needed
    // Typing
    forall (node x) PATS(x->l).
      x!=NULL && tag[x] == node_tag && x->l != NULL ==> tag[x->l] == node_tag;

    forall (node x) PATS(x->r).
      x!=NULL && tag[x] == node_tag && x->r != NULL ==> tag[x->r] == node_tag;
    #endif // 0


    // Global invariant
    forall (node x) PATS(tag[x]).
      x!=NULL && tag[x] == node_tag && x->state==NOT_SEEN ==> x->m == 0;

    forall (node x) PATS(x->l).
      x!=NULL && x->state==NOT_SEEN ==> x->l == x->ol;

    forall (node x) PATS(x->r).
      x!=NULL && x->state==NOT_SEEN ==> x->r == x->or;

    forall (node x) PATS(x->l).
      x!=NULL && x->state==DONELEFT ==>
      (x->m == 1 && x->c == 0 &&
       ((x->l->state == DONELEFT && x->l->ol == x) ||
        (x->l->state == DONERIGHT && x->l->or == x)));

    forall (node x) PATS(x->r).
      x!=NULL && x->state==DONERIGHT ==>
      (x->m == 1 && x->c == 1 &&
       ((x->r->state == DONELEFT && x->r->ol == x) ||
        (x->r->state == DONERIGHT && x->r->or == x)));

    forall (node x) PATS(x->l).
      x!=NULL && x->state==DONERIGHT ==>
      (x->l == x->ol);

    forall (node x) PATS(x->l).
      x!=NULL && x->state==DONE ==>
      (x->m == 1 && x->l == x->ol);
    
    forall (node x) PATS(x->r).
      x!=NULL && x->state==DONE ==>
      (x->m == 1 && x->r == x->or);
    
  });
}

void schorr_waite(node root)
     pre(all_of {
       root != NULL ;
       tag[root] == node_tag;
       forall (node x) PATS(x->state).
         x!=NULL ==> x->state == NOT_SEEN;
     })
     post(all_of {
       forall (node x) PATS(x->state). 
         x!=NULL ==> (x->state == NOT_SEEN || x->state == DONE);

       root->state == DONE;
     })

{
  node t = root;
  node p = NULL;
  while(p != NULL || (t != NULL && t->m != 0)) {

    invariant(all_of {
      1;
    });
    
    node q;
    if(t == NULL || t->m) {
      if(p->c) {
        /* pop */
        q = t;
        t = p;
        p = p->r;
        p->r = p->l;
        p->l = q;
        p->c = 1;
      } else {
        /* swing */
        q = t;
        t = p->r;
        p->r = p->l;
        p->l = q;
        p->c = 1;
      }
    } else {
      /* push */
      q = p;
      p = t;
      t = t->l;
      p->l = q;
      p->m = 1;
      p->c = 0;
    }
  }
}
