// set1.c
// using sets to talk about reachability


// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar
#include "arrayset.model"   // Set, etc.
#include "dpo.model"        // DLT, etc.

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int val, int length);

// temporary: make the PATS go away
//#define PATS(...) /*nothing*/


// --------------- code to verify -----------------
typedef struct Node {
  int data;
  struct Node *next;
  
  // ghost fields
  struct Node *prev;
  int order;
  Set s;               // { x | exists nat n.  this (->next)^n -> data == x }
} Node;

// this node's 's' might not be right
Node *exc = NULL;

// top and bottom order elements
int top_order, bottom_order;


#define NODE_TYPE(val) (0 < (val) < objct && tag[val]==Node_tag)

// can't wrap 'pre' around predicates at the moment...
#define PRE_NODE_TYPE(val) (0 < pre(val) < pre(objct) && pre(tag[val])==Node_tag)


int global_inv() { return vc_expression(all_of {
  objct > 0;
  tag[NULL]==0;
  DLT(bottom_order, top_order);

  // well-typed values
  forall(Node *n) PATS(n->next).
    tag[n]==Node_tag && n->next!=NULL ==>
      NODE_TYPE(n->next);

  forall(Node *n) PATS(n->prev).
    tag[n]==Node_tag && n->prev!=NULL ==>
      NODE_TYPE(n->prev);

  exc!=NULL ==> NODE_TYPE(exc);

  // bottom < order < top
  forall(Node *n) PATS(n->order).                             // order
    tag[n]==Node_tag ==>
      DLT(bottom_order, n->order);
  forall(Node *n) PATS(n->order).                             // order
    tag[n]==Node_tag ==>
      DLT(n->order, top_order);

  // injectivity of next/prev
  forall(Node *n) PATS(n->next).                              // inj
    tag[n]==Node_tag && n->next!=NULL ==>
      n->next->prev == n;
  forall(Node *n) PATS(n->prev).                              // inj
    tag[n]==Node_tag && n->prev!=NULL ==>
      n->prev->next == n;

  // termination condition, makes 's' tight
  forall(Node *n) PATS(tag[n]).                               // neq
    tag[n]==Node_tag && n->next==NULL ==>
      n->s == singletonSet(n->data);

  // transitivity ensuring 's' is big enough
  forall(Node *n) PATS(n->next).                              // trans
    tag[n]==Node_tag && n->next!=NULL && n!=exc ==>
      n->s == singletonUnion(n->data, n->next->s);

  // strict order
  forall(Node *n) PATS(n->next).                              // order
    tag[n]==Node_tag && n->next!=NULL && n!=exc ==>
      isStrictSuperset(n->s, n->next->s);

  // another strict order, for two reasons:
  //   - need it for disequalities, but Kettle doesn't let me declare
  //     arbitrary order symbols
  //   - need the order info across the recursive call to removeElement,
  //     but I modify 's' across that call, so I would probably need
  //     an auxilliary order element anyway
  forall(Node *n) PATS(n->next).                              // order
    tag[n]==Node_tag && n->next!=NULL ==>
      DLT(n->order, n->next->order);
});}


void addElement(Node *head, int data)
  pre(all_of {
    NODE_TYPE(head);
    exc == NULL;
  })
  post(all_of {
    head->s == singletonUnion(data, pre(head->s));
    exc == pre(head->prev);
    objct >= pre(objct);

    // for those objects that had already been allocated, did not
    // modify anything other than the 's' or 'next' fields
    forall(Node *n) PATS(pre(tag[n]), tag[n]).
      PRE_NODE_TYPE(n) ==> tag[n] == pre(tag[n]);
    forall(Node *n) PATS(pre(n->prev), n->prev).
      PRE_NODE_TYPE(n) ==> n->prev == pre(n->prev);
    forall(Node *n) PATS(pre(n->data), n->data).
      PRE_NODE_TYPE(n) ==> n->data == pre(n->data);
    forall(Node *n) PATS(pre(n->order), n->order).
      PRE_NODE_TYPE(n) ==> n->order == pre(n->order);

    // for 's', it will not change if the 'order' is less than 'head->order'
    forall(Node *n) PATS(pre(n->s), n->s).
      PRE_NODE_TYPE(n) && DLT(n->order, pre(head->order)) ==> n->s == pre(n->s);

    // I thought the assertion about 'prev' not changing would be
    // irrelevant, but it turns out to be needed on path 0/2, and
    // I am not sure exactly why...

  })
  changes(objct, tag, Node, exc, objsize)
{
  if (head->data == data) {
    // do nothing
  }

  else if (head->next == NULL) {        // base case of recursion
    // tack on another node
    Node *n = (Node*)malloc(sizeof(Node));
    n->data = data;
    n->next = NULL;
    n->prev = head;
    n->order = DLT_between(head->order, top_order);
    n->s = singletonSet(data);

    head->next = n;
    head->s = singletonUnion(data, head->s);

    // lemma: commutativity of singletonUnion, proved via extensionality
    assert(equalSets(head->s, singletonUnion(head->data, head->next->s)));
  }

  else {                                // recursive case
    addElement(head->next, data);
    assert(exc == head);

    // I can now prove this, by using the subset order
    assert(pre(head->s) == head->s);

    head->s = singletonUnion(data, head->s);
  }

  exc = head->prev;

  // lemma: assist proof of postcondition by appealing to extensionality
  assert(equalSets(head->s, singletonUnion(data, pre(head->s))));
  assert(head->next!=NULL ==>
           equalSets(head->s, singletonUnion(head->data, head->next->s)));
}


Node *removeElement(Node *head, int data)
  pre(all_of {
    NODE_TYPE(head);
    head->prev == NULL;
    containsElement(data, head->s);
    exc == NULL;
  })
  post(all_of {
    result==NULL || NODE_TYPE(result);
    result==pre(head) || result==pre(head->next);
    result!=NULL ==> result->s == removeSingleton(data, pre(head->s));
    result==NULL ==> isEmptySet(removeSingleton(data, pre(head->s)));
    result!=NULL ==> result->prev == NULL;
    Node_order == pre(Node_order);
    
    // things don't change if they are less than 'head'
    forall(Node *n) PATS(pre(tag[n]), tag[n]).
      PRE_NODE_TYPE(n) && DLT(n->order, pre(head->order)) ==> tag[n] == pre(tag[n]);
    forall(Node *n) PATS(pre(n->next), n->next).
      PRE_NODE_TYPE(n) && DLT(n->order, pre(head->order)) ==> n->next == pre(n->next);
    forall(Node *n) PATS(pre(n->prev), n->prev).
      PRE_NODE_TYPE(n) && DLT(n->order, pre(head->order)) ==> n->prev == pre(n->prev);
    forall(Node *n) PATS(pre(n->data), n->data).
      PRE_NODE_TYPE(n) && DLT(n->order, pre(head->order)) ==> n->data == pre(n->data);
  })
  changes(tag, Node)
{
  Node *retval = head;

  if (head->data == data) {       // base case
    retval = head->next;
    if (retval) {
      retval->prev = NULL;
    }

    free(head);
  }
  else {                          // recursive case
    // this is provable from the premise that 'data' is in 'head->s'
    // but is not equal to 'head->data', so 'head->s' is not a
    // singleton set
    assert(NODE_TYPE(head->next));

    // The simplest way to do this recursively is to break the list
    // in half, so that each half respects the list invariants, and
    // then reassemble it afterward.
    //
    // Another way is to allow the invariants to be locally broken,
    // but then I need more exception objects, and things get messy.
    Node *rest = head->next;
    head->next = NULL;
    head->s = singletonSet(head->data);
    rest->prev = NULL;

    // remove 'data' from 'rest'
    rest = removeElement(rest, data);

    // stick 'rest' back onto the list that ends with 'head'
    head->next = rest;
    if (rest) {
      head->s = singletonUnion(head->data, rest->s);
      rest->prev = head;
    }
    else {
      // this assignment is redundant because head->s will not have
      // changed since when it was assigned a few lines above; but
      // proving that is a bit tricky (again, it would be most easily
      // done by using the subset order), and this is a ghost
      // assignment anyway, so I will just do it
      head->s = singletonSet(head->data);
    }
  }

  // lemma
  assert(retval!=NULL ==> equalSets(retval->s, removeSingleton(data, pre(head->s))));

  return retval;
}


// EOF
