; refct.thy
; theory of reference-counted memory objects
; requires: memory.thy

(BG_PUSH (AND

  ; ------------- symbolic constants -------------
  
  ; selection offset where reference count is stored; this is the
  ; count of pointers in the heap or globals
  (EQ refct -2)


  ; ------------- implications of reference counts --------------
  
  (FORALL (mem p1 i1 p2 i2)
    ; try to apply this rule whenever we have two stored values
    (PATS (MPAT
      (sel (sel (selptr mem p1) i1) value)
      (sel (sel (selptr mem p2) i2) value)
    ))

    (IMPLIES
      (AND
        ; given two pointers stored in memory
        ;(EQ (sel (sel (selptr mem p1) i1) value) q1)
        ;(EQ (sel (sel (selptr mem p2) i2) value) q2)
        ; had to substitute q1 and q2 in...
        (isPointer (sel (sel (selptr mem p1) i1) value))  ; q1
        (isPointer (sel (sel (selptr mem p2) i2) value))  ; q2

        ; at different locations
        (OR (NEQ p1 p2) (NEQ i1 i2))

        ; and pointing at reference-count 1 blocks
        (EQ (sel (selptr mem             ; q1
          (sel (sel (selptr mem p1) i1) value)) refct) 1)
        (EQ (sel (selptr mem             ; q2
          (sel (sel (selptr mem p2) i2) value)) refct) 1)
      )

      ; conclude that the stored pointers are different
      (NEQ
        (sel (sel (selptr mem p1) i1) value)    ; q1
        (sel (sel (selptr mem p2) i2) value)    ; q2
      )
    )
  )

  ; ------------- reading/writing reference counts ----------------

  ; (refctOf obj): get obj's refct field
  (FORALL (obj)              (PATS (refctOf obj))
    (EQ
      (refctOf obj)
      (sel obj refct)
    ))

  ; (selRefctPtr mem ptr): get ptr->refct
  (FORALL (mem ptr)          (PATS (selRefctPtr mem ptr))
    (EQ
      (selRefctPtr mem ptr)
      (refctOf (selptr mem ptr))
    ))

  ; (updRefctPtr mem ptr newCt): ptr->refct := newCt
  (FORALL (mem ptr newCt)    (PATS (updRefctPtr mem ptr newCt))
    (EQ
      (updRefctPtr mem ptr newCt)

      ; old = (selptr mem ptr)
      ; new = (upd old refct newCt)
      ; ret = (updptr mem ptr new)
      (updptr mem ptr (upd (selptr mem ptr) refct newCt))
    ))

))
