// invar.c
// kernel showing data structure invariants


// --------- verifier preamble ----------
// for now, use asm hack to put the includes in
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");
asm("verifier: include \"div.vc\";");

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int val, int length);

// declarations of model functions
int sel(int mem, int p);
int ptr(int addr, int field);
int metaVar(int v);

// need to somehow say this is a constant symbolic value, not a global..
// let's try 'const'
int const dead;
int const live;
//int const state;

// more or less true globals
int mem;
int meta;
int tag;
int objsize;
int objct;


// ------------------ code to verify -----------------------
// this structure's "invariants" contribute to the notion
// of validity for instances of the struct
typedef struct Foo {
  int use_flat_struct_model;        // to support the array...
  int x;        // invariant A: 2 <= x <= 5
  int arr[6];   // invariant B: arr[i] is not dead
                // invariant C: arr[x] == 24
} Foo;

// invariants of a Foo at 'addr'; already checked the role tag
int Foo_inv();
define Foo_inv(int mem, int meta, int objct, Foo *addr) {
  0 < addr < objct;
  objsize[addr] == Foo_size;         // size of a Foo
  2 <= addr->x <= 5;
  forall(int i). 0<=i<6 ==> meta[addr,Foo_arr + i]==live;
  //mem[addr,Foo_arr + mem[addr,Foo_x]] == 24;
  mem[addr,Foo_arr + addr->x] == 24;
}


int global_inv() { return vc_expression(all_of {
  0 < objct;
  forall(Foo *addr). tag[addr]==Foo_tag ==>
    Foo_inv(mem, meta, objct, addr);
});}


void anotherFunc(Foo * /*serf*/ p)
  changes(mem, meta)
;

void func()
  changes(mem, meta, tag, objsize, objct)
{
  Foo * /*owner*/ p = (Foo * /*owner*/)malloc(sizeof(Foo));

  // p is an owner pointer to 28 deadbytes

  p->x = 2;

  // invariant A of *p is now satisfied

  for (int i=0; i<6; i++) {
    // invariant A of *p is satisfied
    // for 0<=j<i, p->arr[j] is not dead
//      assert {
//        forall(addr). addr!=p && meta[addr,tag_slot]==Foo_tag ==> (
//          Foo_inv(mem, meta, objct, addr)
//          //( 0 < addr < objct )&&
//          //( meta[addr,state] == Foo_size )&&         // size of a Foo
//          //( 2 <= mem[addr,Foo_x] <= 5 )&&
//          //( forall(i). 0<=i<6 ==> meta[addr,Foo_arr + i]==live )&&
//          //( mem[addr,Foo_arr + mem[addr,Foo_x]] == 24 )&&
//          //1==1
//        );
//      }

    invariant {
      // most of the global invariant holds; the exception is at 'p',
      // where the role tag has been set but the invariant for Foo
      // objects has not been established
      0 < objct;
      forall(Foo *addr). addr!=p && tag[addr]==Foo_tag ==>
        Foo_inv(mem, meta, objct, addr);

      // components of Foo's invariant that have been established
      0 < p < objct;
      tag[p] == Foo_tag;
      objsize[p] == Foo_size;
      2 <= p->x <= 5;

      // invariants of this loop
      metaVar(p) == live;
      0 <= i < 6;
      forall(int j). 0<=j<i ==> meta[p,Foo_arr+j]==live;
    }

    p->arr[i] = i;
  }

  // invariants A and B of *p are now satisfied

  p->arr[p->x] = 24;

  // p is now valid:
  //   p is an owner pointer to Foo-sized memory
  //   all invariants of *p are satisfied
  assert(Foo_inv(mem, meta, objct, p));
  assert(global_inv);

  anotherFunc(p);

  // p is still valid

  free(p);
}

// pre/post: p is valid
void anotherFunc(Foo * /*serf*/ p)
  pre(tag[p] == Foo_tag)     // 'p' is a valid Foo
  changes(mem, meta)
{
  assert mem[p,Foo_arr + mem[p,Foo_x]] == 24;

  if (p->arr[p->x] != 24) {
    assert(0==1);    // ought to be able to prove this is unreachable
  }        
  
  p->x = 5;
  
  // p is not necessarily valid now
  
  p->arr[5] = 24;
  
  // p is valid again
}
