// pc_keyb.c
// single-file implementation of linux pc_keyb.c driver


// --------- verifier preamble ----------
// for now, use asm hack to put the includes in
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");
asm("verifier: include \"div.vc\";");
asm("verifier: include \"bitwise.vc\";");    // bitand

// C functions that are modeled directly
void *malloc(int size);
void kfree(void *freedPtr);

void memset(void *ptr, int val, int length);
void memcpy(void *dest, void const *src, int length);

// declarations of model functions
int sel(int mem, int p);
int ptr(int addr, int field);
int upd(int mem, int addr, int val);
int updflatrange(int mem, int lo, int hi, int v);
int updbothranges(int mem, int loAddr, int hiAddr, int loIndex, int hiIndex, int v);
int canReach(int mem, int field, int src, int dest);
int canReachWO(int mem, int field, int src, int dest, int avoid);
int noCycles(int mem, int field, int ptr);
int disjoint(int mem, int field, int p, int q);
int ptr_field(int ptr);
int ptr_addr(int ptr);
int metaVar(int whatever);
int fresh_value();
int LT(int x, int y);
int LE(int x, int y);

// need to somehow say this is a constant symbolic value, not a global..
// let's try 'const'
int const dead;
int const live;
//int const state;

// more or less true globals
int tag;
int objsize;
int objct;

// aux_queue::buf is modeled as being in this memory, simply because
// I haven't implemented a more sophisticated model yet
int mem;
int meta;
// UPDATE: now I have, so do I need them anymore?  perhaps not.
// but then I want them for read_aux's 'buffer' ..
// TODO: in that case, I don't check bounds on those accesses..

#define AND all_of


/*
 * linux/drivers/char/pc_keyb.c
 *
 * Separation of the PC low-level part by Geert Uytterhoeven, May 1997
 * See keyboard.c for the whole history.
 *
 * Major cleanup by Martin Mares, May 1997
 *
 * Combined the keyboard and PS/2 mouse handling into one file,
 * because they share the same hardware.
 * Johan Myreen <jem@iki.fi> 1998-10-08.
 *
 * Code fixes to handle mouse ACKs properly.
 * C. Scott Ananian <cananian@alumni.princeton.edu> 1999-01-29.
 *
 */

 
// --------------- #include <linux/config.h> ------------------
#define CONFIG_PSMOUSE 1


// unknown source (and might not be correct)
typedef int loff_t;
typedef int ssize_t;
typedef int size_t;
typedef int filldir_t;
typedef int kdev_t;
typedef int pid_t;
typedef unsigned __u32;
typedef int time_t;
#define __init
#define __initdata
#define __setup(a,b)


// linux/kernel.h
#define printk printf
void printf(char const *fmt, ...);
#define	KERN_EMERG	"<0>"	/* system is unusable			*/
#define	KERN_ALERT	"<1>"	/* action must be taken immediately	*/
#define	KERN_CRIT	"<2>"	/* critical conditions			*/
#define	KERN_ERR	"<3>"	/* error conditions			*/
#define	KERN_WARNING	"<4>"	/* warning conditions			*/
#define	KERN_NOTICE	"<5>"	/* normal but significant condition	*/
#define	KERN_INFO	"<6>"	/* informational			*/
#define	KERN_DEBUG	"<7>"	/* debug-level messages			*/


// linux/fs.h
/*
 * NOTE:
 * read, write, poll, fsync, readv, writev can be called
 *   without the big kernel lock held in all filesystems.
 */
struct file_operations {
//	struct module *owner;
	loff_t (*llseek) (struct file *, loff_t, int);
	ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
	int (*readdir) (struct file *, void *, filldir_t);
	unsigned int (*poll) (struct file *, struct poll_table_struct *);
	int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
	int (*mmap) (struct file *, struct vm_area_struct *);
	int (*open) (struct inode *, struct file *);
	int (*flush) (struct file *);
	int (*release) (struct inode *, struct file *);
  	int (*fsync) (struct file *, struct dentry *, int datasync);
  	int (*fasync) (int, struct file *, int);
//  	int (*lock) (struct file *, int, struct file_lock *);
//  	ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *);
//  	ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *);
//  	ssize_t (*sendpage) (struct file *, struct page *, int, size_t, loff_t *, int);
//  	unsigned long (*get_unmapped_area)(struct file *, unsigned long, unsigned long, unsigned long, unsigned long);
};


// linux/fs.h
/* SMP safe fasync helpers: */
extern int fasync_helper(int, struct file *, int, struct fasync_struct **);
/* can be called from interrupts */
extern void kill_fasync(struct fasync_struct **, int, int);


typedef int ModuleId;
enum FileState { FS_CLOSED, FS_OPEN };


// linux/fs.h
struct file {
//  	struct list_head	f_list;
  	struct dentry		*f_dentry;
//  	struct vfsmount         *f_vfsmnt;
	struct file_operations	*f_op;
//  	atomic_t		f_count;
  	unsigned int 		f_flags;
//  	mode_t			f_mode;
  	loff_t			f_pos;
//  	unsigned long 		f_reada, f_ramax, f_raend, f_ralen, f_rawin;
//  	struct fown_struct	f_owner;
//  	unsigned int		f_uid, f_gid;
//  	int			f_error;

//  	unsigned long		f_version;

//  	/* needed for tty driver, and maybe others */
  	void			*private_data;

//  	/* preallocated helper kiobuf to speedup O_DIRECT */
//  	struct kiobuf		*f_iobuf;
//  	long			f_iobuf_lock;

        // annotation (ghost) variables
        ModuleId                id;
        enum FileState               state;
};


// linux/fs.h
struct inode {
//  	struct list_head	i_hash;
//  	struct list_head	i_list;
//  	struct list_head	i_dentry;

//  	struct list_head	i_dirty_buffers;
//  	struct list_head	i_dirty_data_buffers;

//  	unsigned long		i_ino;
//  	atomic_t		i_count;
//  	kdev_t			i_dev;
//  	umode_t			i_mode;
//  	nlink_t			i_nlink;
//  	uid_t			i_uid;
//  	gid_t			i_gid;
  	kdev_t			i_rdev;
//  	loff_t			i_size;
  	time_t			i_atime;
  	time_t			i_mtime;
//  	time_t			i_ctime;
//  	unsigned int		i_blkbits;
//  	unsigned long		i_blksize;
//  	unsigned long		i_blocks;
//  	unsigned long		i_version;
//  	struct semaphore	i_sem;
//  	struct semaphore	i_zombie;
//  	struct inode_operations	*i_op;
//  	struct file_operations	*i_fop;	/* former ->i_op->default_file_ops */
//  	struct super_block	*i_sb;
//  	wait_queue_head_t	i_wait;
//  	struct file_lock	*i_flock;
//  	struct address_space	*i_mapping;
//  	struct address_space	i_data;
//  	struct dquot		*i_dquot[MAXQUOTAS];
//  	/* These three should probably be a union */
//  	struct list_head	i_devices;
//  	struct pipe_inode_info	*i_pipe;
//  	struct block_device	*i_bdev;
//  	struct char_device	*i_cdev;

//  	unsigned long		i_dnotify_mask; /* Directory notify events */
//  	struct dnotify_struct	*i_dnotify; /* for directory notifications */

//  	unsigned long		i_state;

//  	unsigned int		i_flags;
//  	unsigned char		i_sock;

//  	atomic_t		i_writecount;
//  	unsigned int		i_attr_flags;
//  	__u32			i_generation;
//  	union {
//  		struct minix_inode_info		minix_i;
//  		struct ext2_inode_info		ext2_i;
//  		struct ext3_inode_info		ext3_i;
//  		struct hpfs_inode_info		hpfs_i;
//  		struct ntfs_inode_info		ntfs_i;
//  		struct msdos_inode_info		msdos_i;
//  		struct umsdos_inode_info	umsdos_i;
//  		struct iso_inode_info		isofs_i;
//  		struct nfs_inode_info		nfs_i;
//  		struct sysv_inode_info		sysv_i;
//  		struct affs_inode_info		affs_i;
//  		struct ufs_inode_info		ufs_i;
//  		struct efs_inode_info		efs_i;
//  		struct romfs_inode_info		romfs_i;
//  		struct shmem_inode_info		shmem_i;
//  		struct coda_inode_info		coda_i;
//  		struct smb_inode_info		smbfs_i;
//  		struct hfs_inode_info		hfs_i;
//  		struct adfs_inode_info		adfs_i;
//  		struct qnx4_inode_info		qnx4_i;
//  		struct reiserfs_inode_info	reiserfs_i;
//  		struct bfs_inode_info		bfs_i;
//  		struct udf_inode_info		udf_i;
//  		struct ncp_inode_info		ncpfs_i;
//  		struct proc_inode_info		proc_i;
//  		struct socket			socket_i;
//  		struct usbdev_inode_info        usbdev_i;
//  		struct jffs2_inode_info		jffs2_i;
//  		void				*generic_ip;
//  	} u;
};


// asm/errno.h
#define	EIO		 5	/* I/O error */
#define	EAGAIN		11	/* Try again */
#define	ENOMEM		12	/* Out of memory */
#define	EFAULT		14	/* Bad address */
#define	EBUSY		16	/* Device or resource busy */
#define	ENODEV		19	/* No such device */
#define	EINVAL		22	/* Invalid argument */
#define	ERESTART	85	/* Interrupted system call should be restarted */
#define	EUCLEAN		117	/* Structure needs cleaning */

// linux/errno.h
#define ERESTARTSYS	512


// asm/fcntl.h
#define O_ACCMODE	   0003
#define O_RDONLY	     00
#define O_WRONLY	     01
#define O_RDWR		     02
#define O_NONBLOCK	  04000


// linux/dentry.h
struct dentry {
//  	atomic_t d_count;
//  	unsigned int d_flags;
  	struct inode  * d_inode;	/* Where the name belongs to - NULL is negative */
//  	struct dentry * d_parent;	/* parent directory */
//  	struct list_head d_hash;	/* lookup hash list */
//  	struct list_head d_lru;		/* d_count = 0 LRU list */
//  	struct list_head d_child;	/* child of parent list */
//  	struct list_head d_subdirs;	/* our children */
//  	struct list_head d_alias;	/* inode alias list */
//  	int d_mounted;
//  	struct qstr d_name;
//  	unsigned long d_time;		/* used by d_revalidate */
//  	struct dentry_operations  *d_op;
//  	struct super_block * d_sb;	/* The root of the dentry tree */
//  	unsigned long d_vfs_flags;
//  	void * d_fsdata;		/* fs-specific data */
//  	unsigned char d_iname[DNAME_INLINE_LEN]; /* small names */
};



// --------------- #include <linux/spinlock.h> ------------------
                                                  
// not verifying proper operation of these
#define spin_lock_irqsave(lock, flags)          //do { local_irq_save(flags);       spin_lock(lock); } while (0)
#define spin_unlock_irqrestore(lock, flags)     //do { spin_unlock(lock);  local_irq_restore(flags); } while (0)

#define spin_lock_irq(lock)                     //do { local_irq_disable();         spin_lock(lock); } while (0)
#define spin_unlock_irq(lock)			//do { spin_unlock(lock);  local_irq_enable();       } while (0)

// in the single-CPU configuration, spinlock_t is an empty structure..
// for now I'll call it an int, looking forward to actually checking
// that it's used correctly; also I'm not sure my parser can quite
// handle the real definition of SPIN_LOCK_UNLOCKED
typedef int spinlock_t;            // 0=unlocked, 1=locked
#define SPIN_LOCK_UNLOCKED 0

// linux/spinlock.h
#define spin_lock_init(lock)    //do { } while(0)
#define spin_lock(lock)		(void)(lock) /* Not "unused variable". */
#define spin_is_locked(lock)	(0)
#define spin_trylock(lock)	1  //({1; })
#define spin_unlock_wait(lock)  //do { } while(0)
#define spin_unlock(lock)	//do { } while(0)


// linux/wait.h
struct __wait_queue_head {
//  	wq_lock_t lock;
//  	struct list_head task_list;
//  #if WAITQUEUE_DEBUG
//  	long __magic;
//  	long __creator;
//  #endif
};
typedef struct __wait_queue_head wait_queue_head_t;

struct __wait_queue {
//  	unsigned int flags;
//  #define WQ_FLAG_EXCLUSIVE	0x01
//  	struct task_struct * task;
//  	struct list_head task_list;
//  #if WAITQUEUE_DEBUG
//  	long __magic;
//  	long __waker;
//  #endif
};
typedef struct __wait_queue wait_queue_t;

#define DECLARE_WAITQUEUE(name, tsk)					\
        wait_queue_t name;  // = __WAITQUEUE_INITIALIZER(name, tsk)


void init_waitqueue_head(wait_queue_head_t *q);


// --------------- #include <linux/sched.h> ------------------
// linux/sched.h
struct task_struct {
//  	/*
//  	 * offsets of these are hardcoded elsewhere - touch with care
//  	 */
  	volatile long state;	/* -1 unrunnable, 0 runnable, >0 stopped */
//  	unsigned long flags;	/* per process flags, defined below */
//  	int sigpending;
//  	mm_segment_t addr_limit;	/* thread address space:
//  					 	0-0xBFFFFFFF for user-thead
//  						0-0xFFFFFFFF for kernel-thread
//  					 */
//  	struct exec_domain *exec_domain;
//  	volatile long need_resched;
//  	unsigned long ptrace;

//  	int lock_depth;		/* Lock depth */

//  /*
//   * offset 32 begins here on 32-bit platforms. We keep
//   * all fields in a single cacheline that are needed for
//   * the goodness() loop in schedule().
//   */
//  	long counter;
//  	long nice;
//  	unsigned long policy;
//  	struct mm_struct *mm;
//  	int processor;
//  	/*
//  	 * cpus_runnable is ~0 if the process is not running on any
//  	 * CPU. It's (1 << cpu) if it's running on a CPU. This mask
//  	 * is updated under the runqueue lock.
//  	 *
//  	 * To determine whether a process might run on a CPU, this
//  	 * mask is AND-ed with cpus_allowed.
//  	 */
//  	unsigned long cpus_runnable, cpus_allowed;
//  	/*
//  	 * (only the 'next' pointer fits into the cacheline, but
//  	 * that's just fine.)
//  	 */
//  	struct list_head run_list;
//  	unsigned long sleep_time;

//  	struct task_struct *next_task, *prev_task;
//  	struct mm_struct *active_mm;
//  	struct list_head local_pages;
//  	unsigned int allocation_order, nr_local_pages;

//  /* task state */
//  	struct linux_binfmt *binfmt;
//  	int exit_code, exit_signal;
//  	int pdeath_signal;  /*  The signal sent when the parent dies  */
//  	/* ??? */
//  	unsigned long personality;
//  	int did_exec:1;
  	pid_t pid;
//  	pid_t pgrp;
//  	pid_t tty_old_pgrp;
//  	pid_t session;
//  	pid_t tgid;
//  	/* boolean value for session group leader */
//  	int leader;
//  	/* 
//  	 * pointers to (original) parent process, youngest child, younger sibling,
//  	 * older sibling, respectively.  (p->father can be replaced with 
//  	 * p->p_pptr->pid)
//  	 */
//  	struct task_struct *p_opptr, *p_pptr, *p_cptr, *p_ysptr, *p_osptr;
//  	struct list_head thread_group;

//  	/* PID hash table linkage. */
//  	struct task_struct *pidhash_next;
//  	struct task_struct **pidhash_pprev;

//  	wait_queue_head_t wait_chldexit;	/* for wait4() */
//  	struct completion *vfork_done;		/* for vfork() */
//  	unsigned long rt_priority;
//  	unsigned long it_real_value, it_prof_value, it_virt_value;
//  	unsigned long it_real_incr, it_prof_incr, it_virt_incr;
//  	struct timer_list real_timer;
//  	struct tms times;
//  	unsigned long start_time;
//  	long per_cpu_utime[NR_CPUS], per_cpu_stime[NR_CPUS];
//  /* mm fault and swap info: this can arguably be seen as either mm-specific or thread-specific */
//  	unsigned long min_flt, maj_flt, nswap, cmin_flt, cmaj_flt, cnswap;
//  	int swappable:1;
//  /* process credentials */
//  	uid_t uid,euid,suid,fsuid;
//  	gid_t gid,egid,sgid,fsgid;
//  	int ngroups;
//  	gid_t	groups[NGROUPS];
//  	kernel_cap_t   cap_effective, cap_inheritable, cap_permitted;
//  	int keep_capabilities:1;
//  	struct user_struct *user;
//  /* limits */
//  	struct rlimit rlim[RLIM_NLIMITS];
//  	unsigned short used_math;
//  	char comm[16];
//  /* file system info */
//  	int link_count, total_link_count;
//  	struct tty_struct *tty; /* NULL if no tty */
//  	unsigned int locks; /* How many file locks are being held */
//  /* ipc stuff */
//  	struct sem_undo *semundo;
//  	struct sem_queue *semsleeping;
//  /* CPU-specific state of this task */
//  	struct thread_struct thread;
//  /* filesystem information */
//  	struct fs_struct *fs;
//  /* open file information */
//  	struct files_struct *files;
//  /* signal handlers */
//  	spinlock_t sigmask_lock;	/* Protects signal and blocked */
//  	struct signal_struct *sig;

//  	sigset_t blocked;
//  	struct sigpending pending;

//  	unsigned long sas_ss_sp;
//  	size_t sas_ss_size;
//  	int (*notifier)(void *priv);
//  	void *notifier_data;
//  	sigset_t *notifier_mask;
	
//  /* Thread group tracking */
//     	u32 parent_exec_id;
//     	u32 self_exec_id;
//  /* Protection of (de-)allocation: mm, files, fs, tty */
//  	spinlock_t alloc_lock;

//  /* journalling filesystem info */
//  	void *journal_info;
};
struct task_struct *current;    // current process


#define TASK_RUNNING		0
#define TASK_INTERRUPTIBLE	1
extern void __wake_up(wait_queue_head_t *q, unsigned int mode, int nr);
//#define wake_up_interruptible(x)	__wake_up((x),TASK_INTERRUPTIBLE, 1)
#define wake_up_interruptible(x)        //__wake_up((x),TASK_INTERRUPTIBLE, 1)

#define FASTCALL(x) x
extern void FASTCALL(add_wait_queue(wait_queue_head_t *q, wait_queue_t * wait));
//extern void FASTCALL(add_wait_queue_exclusive(wait_queue_head_t *q, wait_queue_t * wait));
extern void FASTCALL(remove_wait_queue(wait_queue_head_t *q, wait_queue_t * wait));


// asm/ptrace.h
struct pt_regs {
	long ebx;
//  	long ecx;
//  	long edx;
//  	long esi;
//  	long edi;
//  	long ebp;
//  	long eax;
//  	int  xds;
//  	int  xes;
//  	long orig_eax;
//  	long eip;
//  	int  xcs;
//  	long eflags;
//  	long esp;
//  	int  xss;
};


extern int request_irq(unsigned int,
		       void (*handler)(int, void *, struct pt_regs *),
		       unsigned long, const char *, void *);
extern void free_irq(unsigned int, void *);


#define __set_current_state(state_value)			\
	do { current->state = (state_value); } while (0)
#ifdef CONFIG_SMP
#define set_current_state(state_value)		\
	set_mb(current->state, (state_value))
#else
#define set_current_state(state_value)		\
	__set_current_state(state_value)
#endif


void schedule(void);


static inline int signal_pending(struct task_struct *p);
//  {
//  	return (p->sigpending != 0);
//  }


int xtime;
#define CURRENT_TIME xtime


// --------------- #include <linux/interrupt.h> ------------------
struct tasklet_struct
{
//  	struct tasklet_struct *next;
//  	unsigned long state;
//  	atomic_t count;
//  	void (*func)(unsigned long);
//  	unsigned long data;
};
void tasklet_schedule(struct tasklet_struct *t);


// --------------- #include <linux/tty.h> ------------------
// --------------- #include <linux/mm.h> ------------------

// --------------- #include <linux/signal.h> ------------------
// asm/signal.h
#define SIGIO		29
#define SA_SHIRQ		0x04000000

// asm/siginfo.h
#define __SI_POLL	(2 << 16)
#define POLL_IN		(__SI_POLL|1)	/* data input available */


// --------------- #include <linux/init.h> ------------------
// --------------- #include <linux/kbd_ll.h> ------------------

// --------------- #include <linux/delay.h> ------------------
#define mdelay(n)     // noop for me, and avoids problems with loops


// --------------- #include <linux/random.h> ------------------
extern void add_mouse_randomness(__u32 mouse_data);


// --------------- #include <linux/poll.h> ------------------
typedef struct poll_table_struct {
//  	int error;
//  	struct poll_table_page * table;
} poll_table;

void poll_wait(struct file * filp, wait_queue_head_t * wait_address, poll_table *p);


// asm/poll.h
#define POLLIN		0x0001
#define POLLRDNORM	0x0040


// --------------- #include <linux/miscdevice.h> ------------------
#define PSMOUSE_MINOR  1

struct miscdevice 
{
	int minor;
	const char *name;
	struct file_operations *fops;
//  	struct miscdevice * next, * prev;
//  	devfs_handle_t devfs_handle;
};

extern int misc_register(struct miscdevice * misc);
extern int misc_deregister(struct miscdevice * misc);


// --------------- #include <linux/slab.h> ------------------
#define kmalloc(sz, mode) malloc(sz)


// --------------- #include <linux/kbd_kern.h> ------------------
extern struct tasklet_struct keyboard_tasklet __varattribute__((addrtaken));


// --------------- #include <linux/vt_kern.h> ------------------
extern int (*kbd_rate)(struct kbd_repeat *rep);


// --------------- #include <linux/smp_lock.h> ------------------
#define lock_kernel()				//do { } while(0)
#define unlock_kernel()				//do { } while(0)


// --------------- #include <linux/kd.h> ------------------
struct kbd_repeat {
	int delay;	/* in msec; <= 0: don't change */
	int rate;	/* in msec; <= 0: don't change */
};


// --------------- #include <linux/pm.h> ------------------
/*
 * Power management requests
 */
enum
{
	PM_SUSPEND, /* enter D1-D3 */
	PM_RESUME,  /* enter D0 */

	PM_SAVE_STATE,  /* save device's state */

	/* enable wake-on */
	PM_SET_WAKEUP,

	/* bus resource management */
	PM_GET_RESOURCES,
	PM_SET_RESOURCES,

	/* base station management */
	PM_EJECT,
	PM_LOCK,
};


typedef int pm_request_t;


// --------------- #include <asm/keyboard.h> ------------------
#define kbd_read_input()       fresh_value() //inb(KBD_DATA_REG)
#define kbd_read_status()      fresh_value() //inb(KBD_STATUS_REG)
#define kbd_write_output(val)  //outb(val, KBD_DATA_REG)
#define kbd_write_command(val) //outb(val, KBD_CNTL_REG)

#define kbd_request_region()
#define kbd_request_irq(handler) request_irq(KEYBOARD_IRQ, handler, 0, \
                                             "keyboard", NULL)

#define KEYBOARD_IRQ			1

#define AUX_IRQ 12
#define aux_request_irq(hand, dev_id)					\
	request_irq(AUX_IRQ, hand, SA_SHIRQ, "PS/2 Mouse", dev_id)
#define aux_free_irq(dev_id) free_irq(AUX_IRQ, dev_id)


// --------------- #include <asm/bitops.h> ------------------

// --------------- #include <asm/uaccess.h> ------------------
#define get_user(var, ptr) (var = *(ptr), 0)
#define put_user(var, ptr) (*(ptr) = var, 0)


// --------------- #include <asm/irq.h> ------------------
// --------------- #include <asm/system.h> ------------------

// --------------- #include <asm/io.h> ------------------


/* Some configuration switches are present in the include file... */
// --------------- #include <linux/pc_keyb.h> ------------------
/*
 *	include/linux/pc_keyb.h
 *
 *	PC Keyboard And Keyboard Controller
 *
 *	(c) 1997 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

/*
 *	Configuration Switches
 */

#undef KBD_REPORT_ERR			/* Report keyboard errors */
#define KBD_REPORT_UNKN			/* Report unknown scan codes */
#define KBD_REPORT_TIMEOUTS		/* Report keyboard timeouts */
#undef KBD_IS_FOCUS_9000		/* We have the brain-damaged FOCUS-9000 keyboard */
#undef INITIALIZE_MOUSE			/* Define if your PS/2 mouse needs initialization. */



#define KBD_INIT_TIMEOUT 1000		/* Timeout in ms for initializing the keyboard */
#define KBC_TIMEOUT 250			/* Timeout in ms for sending to keyboard controller */
#define KBD_TIMEOUT 1000		/* Timeout in ms for keyboard command acknowledge */

/*
 *	Internal variables of the driver
 */

extern unsigned char pckbd_read_mask;
extern unsigned char aux_device_present;

/*
 *	Keyboard Controller Registers on normal PCs.
 */

#define KBD_STATUS_REG		0x64	/* Status register (R) */
#define KBD_CNTL_REG		0x64	/* Controller command register (W) */
#define KBD_DATA_REG		0x60	/* Keyboard data register (R/W) */

/*
 *	Keyboard Controller Commands
 */

#define KBD_CCMD_READ_MODE	0x20	/* Read mode bits */
#define KBD_CCMD_WRITE_MODE	0x60	/* Write mode bits */
#define KBD_CCMD_GET_VERSION	0xA1	/* Get controller version */
#define KBD_CCMD_MOUSE_DISABLE	0xA7	/* Disable mouse interface */
#define KBD_CCMD_MOUSE_ENABLE	0xA8	/* Enable mouse interface */
#define KBD_CCMD_TEST_MOUSE	0xA9	/* Mouse interface test */
#define KBD_CCMD_SELF_TEST	0xAA	/* Controller self test */
#define KBD_CCMD_KBD_TEST	0xAB	/* Keyboard interface test */
#define KBD_CCMD_KBD_DISABLE	0xAD	/* Keyboard interface disable */
#define KBD_CCMD_KBD_ENABLE	0xAE	/* Keyboard interface enable */
#define KBD_CCMD_WRITE_AUX_OBUF	0xD3    /* Write to output buffer as if
					   initiated by the auxiliary device */
#define KBD_CCMD_WRITE_MOUSE	0xD4	/* Write the following byte to the mouse */

/*
 *	Keyboard Commands
 */

#define KBD_CMD_SET_LEDS	0xED	/* Set keyboard leds */
#define KBD_CMD_SET_RATE	0xF3	/* Set typematic rate */
#define KBD_CMD_ENABLE		0xF4	/* Enable scanning */
#define KBD_CMD_DISABLE		0xF5	/* Disable scanning */
#define KBD_CMD_RESET		0xFF	/* Reset */

/*
 *	Keyboard Replies
 */

#define KBD_REPLY_POR		0xAA	/* Power on reset */
#define KBD_REPLY_ACK		0xFA	/* Command ACK */
#define KBD_REPLY_RESEND	0xFE	/* Command NACK, send the cmd again */

/*
 *	Status Register Bits
 */

#define KBD_STAT_OBF 		0x01	/* Keyboard output buffer full */
#define KBD_STAT_IBF 		0x02	/* Keyboard input buffer full */
#define KBD_STAT_SELFTEST	0x04	/* Self test successful */
#define KBD_STAT_CMD		0x08	/* Last write was a command write (0=data) */
#define KBD_STAT_UNLOCKED	0x10	/* Zero if keyboard locked */
#define KBD_STAT_MOUSE_OBF	0x20	/* Mouse output buffer full */
#define KBD_STAT_GTO 		0x40	/* General receive/xmit timeout */
#define KBD_STAT_PERR 		0x80	/* Parity error */

#define AUX_STAT_OBF (KBD_STAT_OBF | KBD_STAT_MOUSE_OBF)

/*
 *	Controller Mode Register Bits
 */

#define KBD_MODE_KBD_INT	0x01	/* Keyboard data generate IRQ1 */
#define KBD_MODE_MOUSE_INT	0x02	/* Mouse data generate IRQ12 */
#define KBD_MODE_SYS 		0x04	/* The system flag (?) */
#define KBD_MODE_NO_KEYLOCK	0x08	/* The keylock doesn't affect the keyboard if set */
#define KBD_MODE_DISABLE_KBD	0x10	/* Disable keyboard interface */
#define KBD_MODE_DISABLE_MOUSE	0x20	/* Disable mouse interface */
#define KBD_MODE_KCC 		0x40	/* Scan code conversion to PC format */
#define KBD_MODE_RFU		0x80

/*
 *	Mouse Commands
 */

#define AUX_SET_RES		0xE8	/* Set resolution */
#define AUX_SET_SCALE11		0xE6	/* Set 1:1 scaling */
#define AUX_SET_SCALE21		0xE7	/* Set 2:1 scaling */
#define AUX_GET_SCALE		0xE9	/* Get scaling factor */
#define AUX_SET_STREAM		0xEA	/* Set stream mode */
#define AUX_SET_SAMPLE		0xF3	/* Set sample rate */
#define AUX_ENABLE_DEV		0xF4	/* Enable aux device */
#define AUX_DISABLE_DEV		0xF5	/* Disable aux device */
#define AUX_RESET		0xFF	/* Reset aux device */
#define AUX_ACK			0xFA	/* Command byte ACK. */

#define AUX_BUF_SIZE		2048	/* This might be better divisible by
					   three to make overruns stay in sync
					   but then the read function would need
					   a lock etc - ick */

struct aux_queue {
	unsigned long head;
	unsigned long tail;
	wait_queue_head_t proc_list;
	struct fasync_struct *fasync;
	unsigned char buf[AUX_BUF_SIZE];
};


int aux_queue_invariants();
define aux_queue_invariants(struct aux_queue *q) {
  tag[q]==aux_queue_tag;

  0 <= q->head < AUX_BUF_SIZE;
  0 <= q->tail < AUX_BUF_SIZE;

  // what about 'proc_list' and 'fasync'?
}



// ---------------------- pc_keyb.c -------------------------
/* Simple translation table for the SysRq keys */

enum ModuleState { MS_UNINIT, MS_INIT } module_state = MS_UNINIT;


#ifdef CONFIG_MAGIC_SYSRQ
unsigned char pckbd_sysrq_xlate[128] =
	"\000\0331234567890-=\177\t"			/* 0x00 - 0x0f */
	"qwertyuiop[]\r\000as"				/* 0x10 - 0x1f */
	"dfghjkl;'`\000\\zxcv"				/* 0x20 - 0x2f */
	"bnm,./\000*\000 \000\201\202\203\204\205"	/* 0x30 - 0x3f */
	"\206\207\210\211\212\000\000789-456+1"		/* 0x40 - 0x4f */
	"230\177\000\000\213\214\000\000\000\000\000\000\000\000\000\000" /* 0x50 - 0x5f */
	"\r\000/";					/* 0x60 - 0x6f */
#endif

static void kbd_write_command_w(int data);
static void kbd_write_output_w(int data);
#ifdef CONFIG_PSMOUSE
static void aux_write_ack(int val);
static void __aux_write_ack(int val);
static int aux_reconnect = 0;
#endif

static spinlock_t kbd_controller_lock = SPIN_LOCK_UNLOCKED;

// moved down
//  static unsigned char handle_kbd_event(void);

/* used only by send_data - set by keyboard_interrupt */
static volatile unsigned char reply_expected;
static volatile unsigned char acknowledge;
static volatile unsigned char resend;


#if defined CONFIG_PSMOUSE
/*
 *	PS/2 Auxiliary Device
 */

static int __init psaux_init(void);

#define AUX_RECONNECT1 0xaa	/* scancode1 when ps2 device is plugged (back) in */
#define AUX_RECONNECT2 0x00	/* scancode2 when ps2 device is plugged (back) in */

static struct aux_queue *queue;	/* Mouse data buffer. */
static int aux_count;
/* used when we send commands to the mouse that expect an ACK. */
static unsigned char mouse_reply_expected;

#define AUX_INTS_OFF (KBD_MODE_KCC | KBD_MODE_DISABLE_MOUSE | KBD_MODE_SYS | KBD_MODE_KBD_INT)
#define AUX_INTS_ON  (KBD_MODE_KCC | KBD_MODE_SYS | KBD_MODE_MOUSE_INT | KBD_MODE_KBD_INT)

#define MAX_RETRIES	60		/* some aux operations take long time*/
#endif /* CONFIG_PSMOUSE */

/*
 * Translation of escaped scancodes to keycodes.
 * This is now user-settable.
 * The keycodes 1-88,96-111,119 are fairly standard, and
 * should probably not be changed - changing might confuse X.
 * X also interprets scancode 0x5d (KEY_Begin).
 *
 * For 1-88 keycode equals scancode.
 */

#define E0_KPENTER 96
#define E0_RCTRL   97
#define E0_KPSLASH 98
#define E0_PRSCR   99
#define E0_RALT    100
#define E0_BREAK   101  /* (control-pause) */
#define E0_HOME    102
#define E0_UP      103
#define E0_PGUP    104
#define E0_LEFT    105
#define E0_RIGHT   106
#define E0_END     107
#define E0_DOWN    108
#define E0_PGDN    109
#define E0_INS     110
#define E0_DEL     111

#define E1_PAUSE   119

/*
 * The keycodes below are randomly located in 89-95,112-118,120-127.
 * They could be thrown away (and all occurrences below replaced by 0),
 * but that would force many users to use the `setkeycodes' utility, where
 * they needed not before. It does not matter that there are duplicates, as
 * long as no duplication occurs for any single keyboard.
 */
#define SC_LIM 89

#define FOCUS_PF1 85           /* actual code! */
#define FOCUS_PF2 89
#define FOCUS_PF3 90
#define FOCUS_PF4 91
#define FOCUS_PF5 92
#define FOCUS_PF6 93
#define FOCUS_PF7 94
#define FOCUS_PF8 95
#define FOCUS_PF9 120
#define FOCUS_PF10 121
#define FOCUS_PF11 122
#define FOCUS_PF12 123

#define JAP_86     124
/* tfj@olivia.ping.dk:
 * The four keys are located over the numeric keypad, and are
 * labelled A1-A4. It's an rc930 keyboard, from
 * Regnecentralen/RC International, Now ICL.
 * Scancodes: 59, 5a, 5b, 5c.
 */
#define RGN1 124
#define RGN2 125
#define RGN3 126
#define RGN4 127

static unsigned char high_keys[128 - SC_LIM] = {
  RGN1, RGN2, RGN3, RGN4, 0, 0, 0,                   /* 0x59-0x5f */
  0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x60-0x67 */
  0, 0, 0, 0, 0, FOCUS_PF11, 0, FOCUS_PF12,          /* 0x68-0x6f */
  0, 0, 0, FOCUS_PF2, FOCUS_PF9, 0, 0, FOCUS_PF3,    /* 0x70-0x77 */
  FOCUS_PF4, FOCUS_PF5, FOCUS_PF6, FOCUS_PF7,        /* 0x78-0x7b */
  FOCUS_PF8, JAP_86, FOCUS_PF10, 0                   /* 0x7c-0x7f */
};

/* BTC */
#define E0_MACRO   112
/* LK450 */
#define E0_F13     113
#define E0_F14     114
#define E0_HELP    115
#define E0_DO      116
#define E0_F17     117
#define E0_KPMINPLUS 118
/*
 * My OmniKey generates e0 4c for  the "OMNI" key and the
 * right alt key does nada. [kkoller@nyx10.cs.du.edu]
 */
#define E0_OK	124
/*
 * New microsoft keyboard is rumoured to have
 * e0 5b (left window button), e0 5c (right window button),
 * e0 5d (menu button). [or: LBANNER, RBANNER, RMENU]
 * [or: Windows_L, Windows_R, TaskMan]
 */
#define E0_MSLW	125
#define E0_MSRW	126
#define E0_MSTM	127

static unsigned char e0_keys[128] = {
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x00-0x07 */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x08-0x0f */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x10-0x17 */
  0, 0, 0, 0, E0_KPENTER, E0_RCTRL, 0, 0,	      /* 0x18-0x1f */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x20-0x27 */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x28-0x2f */
  0, 0, 0, 0, 0, E0_KPSLASH, 0, E0_PRSCR,	      /* 0x30-0x37 */
  E0_RALT, 0, 0, 0, 0, E0_F13, E0_F14, E0_HELP,	      /* 0x38-0x3f */
  E0_DO, E0_F17, 0, 0, 0, 0, E0_BREAK, E0_HOME,	      /* 0x40-0x47 */
  E0_UP, E0_PGUP, 0, E0_LEFT, E0_OK, E0_RIGHT, E0_KPMINPLUS, E0_END,/* 0x48-0x4f */
  E0_DOWN, E0_PGDN, E0_INS, E0_DEL, 0, 0, 0, 0,	      /* 0x50-0x57 */
  0, 0, 0, E0_MSLW, E0_MSRW, E0_MSTM, 0, 0,	      /* 0x58-0x5f */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x60-0x67 */
  0, 0, 0, 0, 0, 0, 0, E0_MACRO,		      /* 0x68-0x6f */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x70-0x77 */
  0, 0, 0, 0, 0, 0, 0, 0			      /* 0x78-0x7f */
};

// globals moved from below
static unsigned char kbd_exists = 1;

#define DEFAULT_KEYB_REP_DELAY	250
#define DEFAULT_KEYB_REP_RATE	30	/* cps */

static struct kbd_repeat kbdrate __varattribute__((addrtaken)) ={
	DEFAULT_KEYB_REP_DELAY,
	DEFAULT_KEYB_REP_RATE
};

/*
 * In case we run on a non-x86 hardware we need to initialize both the
 * keyboard controller and the keyboard.  On a x86, the BIOS will
 * already have initialized them.
 *
 * Some x86 BIOSes do not correctly initialize the keyboard, so the
 * "kbd-reset" command line options can be given to force a reset.
 * [Ranger]
 */
#ifdef __i386__
 int kbd_startup_reset __initdata = 0;
#else
 int kbd_startup_reset __initdata = 1;
#endif


// prototypes for stuff in 'psaux_fops'; at least one of them
// is called explicitly below...
static ssize_t read_aux(struct file * file, char * buffer,
			size_t count, loff_t *ppos);
static ssize_t write_aux(struct file * file, const char * buffer,
			 size_t count, loff_t *ppos);
static unsigned int aux_poll(struct file *file, poll_table * wait);
static int open_aux(struct inode * inode, struct file * file);
static int release_aux(struct inode * inode, struct file * file);
static int fasync_aux(int fd, struct file *filp, int on);



struct file_operations psaux_fops = {
	/*llseek:*/     NULL,
	/*read:*/       read_aux,
	/*write:*/      write_aux,
	/*readder:*/    NULL,
	/*poll:*/       aux_poll,
	/*ioctl:*/      NULL,
	/*mmap:*/       NULL,
        /*open:*/       open_aux,
        /*flush:*/      NULL,
        /*release:*/    release_aux,
        /*fsync:*/      NULL,
	/*fasync:*/     fasync_aux,
};

/*
 * Initialize driver.
 */
static struct miscdevice psaux_mouse = {
	PSMOUSE_MINOR, "psaux", &psaux_fops
};


// moved up from parse_kbd_rate
struct r2v{
  int rate;
  unsigned char val;
} kbd_rates[]={	{5,0x14},
                {7,0x10},
                {10,0x0c},
                {15,0x08},
                {20,0x04},
                {25,0x02},
                {30,0x00}
};

struct d2v{
  int delay;
  unsigned char val;
} kbd_delays[]={{250,0},
                {500,1},
                {750,2},
                {1000,3}
};


// moved up from pckbd_rate
struct kbd_repeat old_rep __varattribute__((addrtaken));


// model the address of some functions
//int pckbd_rate_address;
//int keyboard_interrupt_address;


// TODO: need to check that the global invariant holds for the
// initial values of all the global variables
int global_inv() { return vc_expression(all_of {
  module_state==MS_UNINIT || module_state==MS_INIT;
  module_state==MS_INIT ==> all_of {
    0 < queue < objct;
    aux_queue_invariants(queue);
  };

  0 < current < objct && tag[current]==task_struct_tag;

  forall(struct file *f) PATS(f->f_dentry). tag[f]==file_tag ==> AND {
    0 < f->f_dentry < objct && tag[f->f_dentry]==dentry_tag;                // disequality
  };

  forall(struct dentry *d) PATS(d->d_inode). tag[d]==dentry_tag ==> AND {
    0 < d->d_inode < objct && tag[d->d_inode]==inode_tag;                   // disequality
  };

  // all the entries in e0_keys must be < 128
  // TODO: allow 'sizeof', etc in VC exprs
  forall(int i). 0<=i<128 ==> 0<=e0_keys[i]<128;

  // similar for entries in high_keys
  forall(int i). 0<=i<(128-SC_LIM) ==> 0<=high_keys[i]<128;

  // kbd_rates and kbd_rates
  forall(int i). 0<=i<7 ==> tag[kbd_rates+i]==r2v_tag;
  forall(int i). 0<=i<4 ==> tag[kbd_delays+i]==d2v_tag;
});}


// things that reach handle_mouse_event change these variables
change_group cg_handle_mouse_event =
  mouse_reply_expected, /*mem, meta,*/ aux_queue_head, aux_queue_tail,
  aux_queue_buf, acknowledge, reply_expected, resend;


static unsigned char handle_kbd_event(void)
  // can be called in either state
  changes(cg_handle_mouse_event)
;


/*
 * Wait for keyboard controller input buffer to drain.
 *
 * Don't use 'jiffies' so that we don't depend on
 * interrupts..
 *
 * Quote from PS/2 System Reference Manual:
 *
 * "Address hex 0060 and address hex 0064 should be written only when
 * the input-buffer-full bit and output-buffer-full bit in the
 * Controller Status register are set 0."
 */

static void kb_wait(void)
  // can be called in either state
  changes(cg_handle_mouse_event)
{
	unsigned long timeout = KBC_TIMEOUT;

	do {
		invariant(global_inv);

		/*
		 * "handle_kbd_event()" will handle any incoming events
		 * while we wait - keypresses or mouse movement.
		 */
		unsigned char status = handle_kbd_event();

		if (! (status & KBD_STAT_IBF))
			return;
		mdelay(1);
		timeout--;
	} while (timeout);
#ifdef KBD_REPORT_TIMEOUTS
	printk(KERN_WARNING "Keyboard timed out[1]\n");
#endif
}

int pckbd_setkeycode(unsigned int scancode, unsigned int keycode)
  changes(high_keys, e0_keys)
{
	if (scancode < SC_LIM || scancode > 255 || keycode > 127)
	  return -EINVAL;
	if (scancode < 128)
	  high_keys[scancode - SC_LIM] = keycode;
	else
	  e0_keys[scancode - 128] = keycode;
	return 0;
}

int pckbd_getkeycode(unsigned int scancode)
{
	return
	  (scancode < SC_LIM || scancode > 255) ? -EINVAL :
	  (scancode < 128) ? high_keys[scancode - SC_LIM] :
	    e0_keys[scancode - 128];
}


change_group cg_do_acknowledge = acknowledge, reply_expected, resend;

static int do_acknowledge(unsigned char scancode)
  changes(cg_do_acknowledge)
{
	if (reply_expected) {
	  /* Unfortunately, we must recognise these codes only if we know they
	   * are known to be valid (i.e., after sending a command), because there
	   * are some brain-damaged keyboards (yes, FOCUS 9000 again) which have
	   * keys with such codes :(
	   */
		if (scancode == KBD_REPLY_ACK) {
			acknowledge = 1;
			reply_expected = 0;
			return 0;
		} else if (scancode == KBD_REPLY_RESEND) {
			resend = 1;
			reply_expected = 0;
			return 0;
		}
		/* Should not happen... */
#if 0
		printk(KERN_DEBUG "keyboard reply expected - got %02x\n",
		       scancode);
#endif
	}
	return 1;
}

// 'keycode' is a call-by reference extra return value; example call
// site at /usr/src/linux/drivers/char/keyboard.c, line 232:
//   unsigned char keycode;
//   if (!kbd_translate(scancode, &keycode, raw_mode))
// furthermore, if this function returns nonzero, then the final
// value of 'keycode' must be within [0,127] b/c it's used to
// index the sysrq table
int pckbd_translate(unsigned char scancode,
                    unsigned char *keycode __varattribute__((parambyref)),
		    char raw_mode)
  post(result!=0 ==> (0 <= *keycode <= 127))
{
	static int prev_scancode;

	/* special prefix scancodes.. */
	if (scancode == 0xe0 || scancode == 0xe1) {
		prev_scancode = scancode;
		return 0;
	}

	/* 0xFF is sent by a few keyboards, ignore it. 0x00 is error */
	if (scancode == 0x00 || scancode == 0xff) {
		prev_scancode = 0;
		return 0;
	}

	scancode &= 0x7f;

	// lemma: arithmetic consequence of bitwise operation
	assert LE(0, scancode) && LT(scancode, 0x80);

	if (prev_scancode) {
	  /*
	   * usually it will be 0xe0, but a Pause key generates
	   * e1 1d 45 e1 9d c5 when pressed, and nothing when released
	   */
	  if (prev_scancode != 0xe0) {
	      if (prev_scancode == 0xe1 && scancode == 0x1d) {
		  prev_scancode = 0x100;
		  return 0;
	      } else if (prev_scancode == 0x100 && scancode == 0x45) {
		  *keycode = E1_PAUSE;
		  prev_scancode = 0;
	      } else {
#ifdef KBD_REPORT_UNKN
		  if (!raw_mode)
		    printk(KERN_INFO "keyboard: unknown e1 escape sequence\n");
#endif
		  prev_scancode = 0;
		  return 0;
	      }
	  } else {
	      prev_scancode = 0;
	      /*
	       *  The keyboard maintains its own internal caps lock and
	       *  num lock statuses. In caps lock mode E0 AA precedes make
	       *  code and E0 2A follows break code. In num lock mode,
	       *  E0 2A precedes make code and E0 AA follows break code.
	       *  We do our own book-keeping, so we will just ignore these.
	       */
	      /*
	       *  For my keyboard there is no caps lock mode, but there are
	       *  both Shift-L and Shift-R modes. The former mode generates
	       *  E0 2A / E0 AA pairs, the latter E0 B6 / E0 36 pairs.
	       *  So, we should also ignore the latter. - aeb@cwi.nl
	       */
	      if (scancode == 0x2a || scancode == 0x36)
		return 0;

	      if (e0_keys[scancode])
		*keycode = e0_keys[scancode];
	      else {
#ifdef KBD_REPORT_UNKN
		  if (!raw_mode)
		    printk(KERN_INFO "keyboard: unknown scancode e0 %02x\n",
			   scancode);
#endif
		  return 0;
	      }
	  }
	} else if (scancode >= SC_LIM) {
	    /* This happens with the FOCUS 9000 keyboard
	       Its keys PF1..PF12 are reported to generate
	       55 73 77 78 79 7a 7b 7c 74 7e 6d 6f
	       Moreover, unless repeated, they do not generate
	       key-down events, so we have to zero up_flag below */
	    /* Also, Japanese 86/106 keyboards are reported to
	       generate 0x73 and 0x7d for \ - and \ | respectively. */
	    /* Also, some Brazilian keyboard is reported to produce
	       0x73 and 0x7e for \ ? and KP-dot, respectively. */

	  *keycode = high_keys[scancode - SC_LIM];

	  if (!*keycode) {
	      if (!raw_mode) {
#ifdef KBD_REPORT_UNKN
		  printk(KERN_INFO "keyboard: unrecognized scancode (%02x)"
			 " - ignored\n", scancode);
#endif
	      }
	      return 0;
	  }
 	} else
	  *keycode = scancode;
 	return 1;
}

char pckbd_unexpected_up(unsigned char keycode)
{
	/* unexpected, but this can happen: maybe this was a key release for a
	   FOCUS 9000 PF key; if we want to see it, we have to clear up_flag */
	if (keycode >= SC_LIM || keycode == 85)
	    return 0;
	else
	    return 0200;
}

int pckbd_pm_resume(struct pm_dev *dev, pm_request_t rqst, void *data)
  pre(module_state==MS_INIT)
  changes(cg_handle_mouse_event)
{
#if defined CONFIG_PSMOUSE
       unsigned long flags;

       if (rqst == PM_RESUME) {
               if (queue) {                    /* Aux port detected */
                       if (aux_count == 0) {   /* Mouse not in use */
                               spin_lock_irqsave(&kbd_controller_lock, flags);
			       /*
				* Dell Lat. C600 A06 enables mouse after resume.
				* When user touches the pad, it posts IRQ 12
				* (which we do not process), thus holding keyboard.
				*/
			       kbd_write_command(KBD_CCMD_MOUSE_DISABLE);
			       /* kbd_write_cmd(AUX_INTS_OFF); */ /* Config & lock */
			       kb_wait();
			       kbd_write_command(KBD_CCMD_WRITE_MODE);
			       kb_wait();
			       kbd_write_output(AUX_INTS_OFF);
			       spin_unlock_irqrestore(&kbd_controller_lock, flags);
		       }
	       }
       }
#endif
       return 0;
}


static inline void handle_mouse_event(unsigned char scancode)
  pre(module_state==MS_INIT)
  changes(cg_handle_mouse_event)
{
#ifdef CONFIG_PSMOUSE
	static unsigned char prev_code;
	if (mouse_reply_expected) {
		if (scancode == AUX_ACK) {
			mouse_reply_expected--;
			return;
		}
		mouse_reply_expected = 0;
	}
	else if(scancode == AUX_RECONNECT2 && prev_code == AUX_RECONNECT1
		&& aux_reconnect) {
		printk (KERN_INFO "PS/2 mouse reconnect detected\n");
		queue->head = queue->tail = 0;	/* Flush input queue */
		__aux_write_ack(AUX_ENABLE_DEV);  /* ping the mouse :) */
		return;
	}

	prev_code = scancode;
	add_mouse_randomness(scancode);
	if (aux_count) {
		int head = queue->head;

		queue->buf[head] = scancode;
		head = (head + 1) & (AUX_BUF_SIZE-1);
		assert LE(0,head) && LT(head,AUX_BUF_SIZE);     // consequence of bitwise AND
		if (head != queue->tail) {
			queue->head = head;
			kill_fasync(&queue->fasync, SIGIO, POLL_IN);
			wake_up_interruptible(&queue->proc_list);
		}
	}
#endif
}
  
// moved up
//static unsigned char kbd_exists = 1;

static inline void handle_keyboard_event(unsigned char scancode)
  pre(module_state==MS_INIT)
  changes(cg_do_acknowledge)
{
#ifdef CONFIG_VT
	kbd_exists = 1;
	if (do_acknowledge(scancode))
		handle_scancode(scancode, !(scancode & 0x80));
#endif
	tasklet_schedule(&keyboard_tasklet);
}

/*
 * This reads the keyboard status port, and does the
 * appropriate action.
 *
 * It requires that we hold the keyboard controller
 * spinlock.
 */
static unsigned char handle_kbd_event(void)
  // can be called in either state
  changes(cg_handle_mouse_event)
{
	unsigned char status = kbd_read_status();
	unsigned int work = 10000;

	// sm: This function is called while the module is not
	// initialized, but 'handle_mouse_event' certainly requires
	// that the module *be* initialized.  Therefore I must assume
	// that 'kbd_read_status' does not return KBD_STAT_OBF when
	// the module is uninitialized.  Perhaps this means there is a
	// bug, if the user types on the keyboard early during boot?
	assume(module_state==MS_UNINIT ==> status == 0);

	while ((--work > 0) && (status & KBD_STAT_OBF)) {
		invariant {
		  global_inv;
		  module_state==MS_INIT;
		}

		unsigned char scancode;

		scancode = kbd_read_input();

		/* Error bytes must be ignored to make the
		   Synaptics touchpads compaq use work */
#if 1
		/* Ignore error bytes */
		if (!(status & (KBD_STAT_GTO | KBD_STAT_PERR)))
#endif
		{
			if (status & KBD_STAT_MOUSE_OBF)
                                handle_mouse_event(scancode);
			else
                                handle_keyboard_event(scancode);
		}

		status = kbd_read_status();
	}
		
	if (!work)
		printk(KERN_ERR "pc_keyb: controller jammed (0x%02X).\n", status);

	return status;
}


static void keyboard_interrupt(int irq, void *dev_id, struct pt_regs *regs)
  pre(module_state==MS_INIT)
  changes(cg_handle_mouse_event)
{
#ifdef CONFIG_VT
	kbd_pt_regs = regs;
#endif

	spin_lock_irq(&kbd_controller_lock);
	handle_kbd_event();
	spin_unlock_irq(&kbd_controller_lock);
}


change_group cg_send_data =
  reply_expected, acknowledge, resend;
/*
 * send_data sends a character to the keyboard and waits
 * for an acknowledge, possibly retrying if asked to. Returns
 * the success status.
 *
 * Don't use 'jiffies', so that we don't depend on interrupts
 */
static int send_data(unsigned char data)
  changes(cg_send_data)
{
	int retries = 3;

	do {
		invariant {
		  global_inv;
		}

		unsigned long timeout = KBD_TIMEOUT;

		acknowledge = 0; /* Set by interrupt routine on receipt of ACK. */
		resend = 0;
                reply_expected = 1;
	        kbd_write_output_w(data);
		for (;;) {
			invariant { global_inv; }
			if (acknowledge)
				return 1;
			if (resend)
				break;
			mdelay(1);
			if (!--timeout) {
#ifdef KBD_REPORT_TIMEOUTS
				printk(KERN_WARNING "keyboard: Timeout - AT keyboard not present?(%02x)\n", data);
#endif
				return 0;
			}
		}
	} while (retries-- > 0);
#ifdef KBD_REPORT_TIMEOUTS
	printk(KERN_WARNING "keyboard: Too many NACKs -- noisy kbd cable?\n");
#endif
	return 0;
}

void pckbd_leds(unsigned char leds)
  changes(kbd_exists, cg_send_data)
{
	if (kbd_exists && (!send_data(KBD_CMD_SET_LEDS) || !send_data(leds))) {
		send_data(KBD_CMD_ENABLE);	/* re-enable kbd if any errors */
		kbd_exists = 0;
	}
}


static unsigned char parse_kbd_rate(struct kbd_repeat *r)
  pre(0 < r < objct && tag[r]==kbd_repeat_tag)
  changes(kbd_repeat_tag)
{
	#if 0   // moved up
	static struct r2v {
		int rate;
	        unsigned char val;
        } kbd_rates[]={	{5,0x14},
	        	{7,0x10},
	        	{10,0x0c},
	        	{15,0x08},
	        	{20,0x04},
	        	{25,0x02},
	        	{30,0x00}
	};
	static struct d2v {
	        int delay;
	        unsigned char val;
	} kbd_delays[]={{250,0},
			{500,1},
			{750,2},
			{1000,3}
	};
	#endif // 0
	int rate=0,delay=0;
	if (r != NULL){
		int i,new_rate=30,new_delay=250;
		if (r->rate <= 0)
			r->rate=kbdrate.rate;
		if (r->delay <= 0)
			r->delay=kbdrate.delay;
		for (i=0; i < sizeof(kbd_rates)/sizeof(struct r2v); i++) {
		        invariant {
		          global_inv;
                          0 < r < objct && tag[r]==kbd_repeat_tag;
		          0 <= i < 7;
		        }
			if (kbd_rates[i].rate == r->rate){
				new_rate=kbd_rates[i].rate;
				rate=kbd_rates[i].val;
				break;
			}
		}
		for (i=0; i < sizeof(kbd_delays)/sizeof(struct d2v); i++) {
		        invariant {
		          global_inv;
                          0 < r < objct && tag[r]==kbd_repeat_tag;
		          0 <= i < 4;
		        }
			if (kbd_delays[i].delay == r->delay){
				new_delay=kbd_delays[i].delay;
				delay=kbd_delays[i].val;
				break;
			}
		}
		r->rate=new_rate;
		r->delay=new_delay;
	}
	return (delay << 5) | rate;
}


static int write_kbd_rate(unsigned char r)
  changes(cg_send_data)
{
	if (!send_data(KBD_CMD_SET_RATE) || !send_data(r)){
		send_data(KBD_CMD_ENABLE); 	/* re-enable kbd if any errors */
		return 0;
	}else
		return 1;
}

// TODO: memcpy
static int pckbd_rate(struct kbd_repeat *rep)
  pre(rep!=NULL ==> 0 < rep < objct && tag[rep]==kbd_repeat_tag)
  changes(kbd_repeat_tag, cg_send_data)
{
	if (rep == NULL)
		return -EINVAL;
	else{
		unsigned char r=parse_kbd_rate(rep);
		//struct kbd_repeat old_rep;
		memcpy(&old_rep,&kbdrate,sizeof(struct kbd_repeat));
		if (write_kbd_rate(r)){
			memcpy(&kbdrate,rep,sizeof(struct kbd_repeat));
			memcpy(rep,&old_rep,sizeof(struct kbd_repeat));
			return 0;
		}
	}
	return -EIO;
}

/* for "kbd-reset" cmdline param */
static int __init kbd_reset_setup(char *str)
  changes(kbd_startup_reset)
{
	kbd_startup_reset = 1;
	return 1;
}

__setup("kbd-reset", kbd_reset_setup);

#define KBD_NO_DATA	(-1)	/* No data */
#define KBD_BAD_DATA	(-2)	/* Parity or other error */

static int __init kbd_read_data(void)
{
	int retval = KBD_NO_DATA;
	unsigned char status;

	status = kbd_read_status();
	if (status & KBD_STAT_OBF) {
		unsigned char data = kbd_read_input();

		retval = data;
		if (status & (KBD_STAT_GTO | KBD_STAT_PERR))
			retval = KBD_BAD_DATA;
	}
	return retval;
}

static void __init kbd_clear_input(void)
{
	int maxread = 100;	/* Random number */

	do {
		invariant { global_inv; }
		if (kbd_read_data() == KBD_NO_DATA)
			break;
	} while (--maxread);
}

static int __init kbd_wait_for_input(void)
{
	long timeout = KBD_INIT_TIMEOUT;

	do {
		invariant { global_inv; }
		int retval = kbd_read_data();
		if (retval >= 0)
			return retval;
		mdelay(1);
	} while (--timeout);
	return -1;
}

static void kbd_write_command_w(int data)
  pre(module_state==MS_INIT)
  changes(cg_handle_mouse_event)
{
	unsigned long flags;

	spin_lock_irqsave(&kbd_controller_lock, flags);
	kb_wait();
	kbd_write_command(data);
	spin_unlock_irqrestore(&kbd_controller_lock, flags);
}

static void kbd_write_output_w(int data)
  pre(module_state==MS_INIT)
  changes(cg_handle_mouse_event)
{
	unsigned long flags;

	spin_lock_irqsave(&kbd_controller_lock, flags);
        kb_wait();
	kbd_write_output(data);
	spin_unlock_irqrestore(&kbd_controller_lock, flags);
}

#if defined(__alpha__)
/*
 * Some Alphas cannot mask some/all interrupts, so we have to
 * make sure not to allow interrupts AT ALL when polling for
 * specific return values from the keyboard.
 *
 * I think this should work on any architecture, but for now, only Alpha.
 */
static int kbd_write_command_w_and_wait(int data)
  changes(cg_handle_mouse_event)
{
	unsigned long flags;
	int input;

	spin_lock_irqsave(&kbd_controller_lock, flags);
	kb_wait();
	kbd_write_command(data);
	input = kbd_wait_for_input();
	spin_unlock_irqrestore(&kbd_controller_lock, flags);
	return input;
}

static int kbd_write_output_w_and_wait(int data)
  changes(cg_handle_mouse_event)
{
	unsigned long flags;
	int input;

	spin_lock_irqsave(&kbd_controller_lock, flags);
	kb_wait();
	kbd_write_output(data);
	input = kbd_wait_for_input();
	spin_unlock_irqrestore(&kbd_controller_lock, flags);
	return input;
}
#else
static int kbd_write_command_w_and_wait(int data)
{
	kbd_write_command_w(data);
	return kbd_wait_for_input();
}

static int kbd_write_output_w_and_wait(int data)
{
	kbd_write_output_w(data);
	return kbd_wait_for_input();
}
#endif /* __alpha__ */

#if defined CONFIG_PSMOUSE
static void kbd_write_cmd(int cmd)
  pre(module_state==MS_INIT)
  changes(cg_handle_mouse_event)
{
	unsigned long flags;

	spin_lock_irqsave(&kbd_controller_lock, flags);
	kb_wait();
	kbd_write_command(KBD_CCMD_WRITE_MODE);
	kb_wait();
	kbd_write_output(cmd);
	spin_unlock_irqrestore(&kbd_controller_lock, flags);
}
#endif /* CONFIG_PSMOUSE */

static char * __init initialize_kbd(void)
{
	int status;

	/*
	 * Test the keyboard interface.
	 * This seems to be the only way to get it going.
	 * If the test is successful a x55 is placed in the input buffer.
	 */
	kbd_write_command_w(KBD_CCMD_SELF_TEST);
	if (kbd_wait_for_input() != 0x55)
		return "Keyboard failed self test";

	/*
	 * Perform a keyboard interface test.  This causes the controller
	 * to test the keyboard clock and data lines.  The results of the
	 * test are placed in the input buffer.
	 */
	kbd_write_command_w(KBD_CCMD_KBD_TEST);
	if (kbd_wait_for_input() != 0x00)
		return "Keyboard interface failed self test";

	/*
	 * Enable the keyboard by allowing the keyboard clock to run.
	 */
	kbd_write_command_w(KBD_CCMD_KBD_ENABLE);

	/*
	 * Reset keyboard. If the read times out
	 * then the assumption is that no keyboard is
	 * plugged into the machine.
	 * This defaults the keyboard to scan-code set 2.
	 *
	 * Set up to try again if the keyboard asks for RESEND.
	 */
	do {
		invariant { global_inv; }
		kbd_write_output_w(KBD_CMD_RESET);
		status = kbd_wait_for_input();
		if (status == KBD_REPLY_ACK)
			break;
		if (status != KBD_REPLY_RESEND)
			return "Keyboard reset failed, no ACK";
	} while (1);

	if (kbd_wait_for_input() != KBD_REPLY_POR)
		return "Keyboard reset failed, no POR";

	/*
	 * Set keyboard controller mode. During this, the keyboard should be
	 * in the disabled state.
	 *
	 * Set up to try again if the keyboard asks for RESEND.
	 */
	do {
		invariant { global_inv; }
		kbd_write_output_w(KBD_CMD_DISABLE);
		status = kbd_wait_for_input();
		if (status == KBD_REPLY_ACK)
			break;
		if (status != KBD_REPLY_RESEND)
			return "Disable keyboard: no ACK";
	} while (1);

	kbd_write_command_w(KBD_CCMD_WRITE_MODE);
	kbd_write_output_w(KBD_MODE_KBD_INT
			      | KBD_MODE_SYS
			      | KBD_MODE_DISABLE_MOUSE
			      | KBD_MODE_KCC);

	/* ibm powerpc portables need this to use scan-code set 1 -- Cort */
	if (!(kbd_write_command_w_and_wait(KBD_CCMD_READ_MODE) & KBD_MODE_KCC))
	{
		/*
		 * If the controller does not support conversion,
		 * Set the keyboard to scan-code set 1.
		 */
		kbd_write_output_w(0xF0);
		kbd_wait_for_input();
		kbd_write_output_w(0x01);
		kbd_wait_for_input();
	}

	if (kbd_write_output_w_and_wait(KBD_CMD_ENABLE) != KBD_REPLY_ACK)
		return "Enable keyboard: no ACK";

	/*
	 * Finally, set the typematic rate to maximum.
	 */
	if (kbd_write_output_w_and_wait(KBD_CMD_SET_RATE) != KBD_REPLY_ACK)
		return "Set rate: no ACK";
	if (kbd_write_output_w_and_wait(0x00) != KBD_REPLY_ACK)
		return "Set rate: no 2nd ACK";

	return NULL;
}

void __init pckbd_init_hw(void)
  changes(kbd_rate)
{
	kbd_request_region();

	/* Flush any pending input. */
	kbd_clear_input();

	if (kbd_startup_reset) {
		char *msg = initialize_kbd();
		if (msg)
			printk(KERN_WARNING "initialize_kbd: %s\n", msg);
	}

#if defined CONFIG_PSMOUSE
	psaux_init();
#endif

	kbd_rate = pckbd_rate;

	/* Ok, finally allocate the IRQ, and off we go.. */
	kbd_request_irq(keyboard_interrupt);
}

#if defined CONFIG_PSMOUSE

static int __init aux_reconnect_setup (char *str)
  changes(aux_reconnect)
{
	aux_reconnect = 1;
	return 1;
}

__setup("psaux-reconnect", aux_reconnect_setup);

/*
 * Check if this is a dual port controller.
 */
static int __init detect_auxiliary_port(void)
  pre(module_state==MS_UNINIT)
  changes(cg_handle_mouse_event)
{
	unsigned long flags;
	int loops = 10;
	int retval = 0;

	/* Check if the BIOS detected a device on the auxiliary port. */
	if (aux_device_present == 0xaa)
		return 1;

	spin_lock_irqsave(&kbd_controller_lock, flags);

	/* Put the value 0x5A in the output buffer using the "Write
	 * Auxiliary Device Output Buffer" command (0xD3). Poll the
	 * Status Register for a while to see if the value really
	 * turns up in the Data Register. If the KBD_STAT_MOUSE_OBF
	 * bit is also set to 1 in the Status Register, we assume this
	 * controller has an Auxiliary Port (a.k.a. Mouse Port).
	 */
        kb_wait();
	kbd_write_command(KBD_CCMD_WRITE_AUX_OBUF);

	kb_wait();
	kbd_write_output(0x5a); /* 0x5a is a random dummy value. */

	do {
		invariant { global_inv; }
		unsigned char status = kbd_read_status();

		if (status & KBD_STAT_OBF) {
			(void) kbd_read_input();
			if (status & KBD_STAT_MOUSE_OBF) {
				printk(KERN_INFO "Detected PS/2 Mouse Port.\n");
				retval = 1;
			}
			break;
		}
		mdelay(1);
	} while (--loops);
	spin_unlock_irqrestore(&kbd_controller_lock, flags);

	return retval;
}

/*
 * Send a byte to the mouse.
 */
static void aux_write_dev(int val)
  pre(module_state==MS_INIT)
  changes(cg_handle_mouse_event)
{
	unsigned long flags;

	spin_lock_irqsave(&kbd_controller_lock, flags);
	kb_wait();
	kbd_write_command(KBD_CCMD_WRITE_MOUSE);
	kb_wait();
	kbd_write_output(val);
	spin_unlock_irqrestore(&kbd_controller_lock, flags);
}

/*
 * Send a byte to the mouse & handle returned ack
 */
static void __aux_write_ack(int val)
  pre(module_state==MS_INIT)
  changes(cg_handle_mouse_event)
{
	kb_wait();
	kbd_write_command(KBD_CCMD_WRITE_MOUSE);
	kb_wait();
	kbd_write_output(val);
	/* we expect an ACK in response. */
	mouse_reply_expected++;
	kb_wait();
}

static void aux_write_ack(int val)
{
	unsigned long flags;

	spin_lock_irqsave(&kbd_controller_lock, flags);
	__aux_write_ack(val);
	spin_unlock_irqrestore(&kbd_controller_lock, flags);
}

static unsigned char get_from_queue(void)
  pre(module_state == MS_INIT)
  changes(aux_queue_tail)
{
	unsigned char resULT;
	unsigned long flags;

	spin_lock_irqsave(&kbd_controller_lock, flags);
	resULT = queue->buf[queue->tail];
	queue->tail = (queue->tail + 1) & (AUX_BUF_SIZE-1);
	assert LE(0, queue->tail) && LT(queue->tail, AUX_BUF_SIZE);
	spin_unlock_irqrestore(&kbd_controller_lock, flags);
	return resULT;
}


static inline int queue_empty(void)
  pre(module_state == MS_INIT)
{
	return queue->head == queue->tail;
}

static int fasync_aux(int fd, struct file *filp, int on)
  pre(module_state == MS_INIT)
{
	int retval;

	retval = fasync_helper(fd, filp, on, &queue->fasync);
	if (retval < 0)
		return retval;
	return 0;
}


/*
 * Random magic cookie for the aux device
 */
#define AUX_DEV ((void *)queue)

static int release_aux(struct inode * inode, struct file * file)
  pre(module_state == MS_INIT)
  changes(aux_count, cg_handle_mouse_event)
{
	lock_kernel();
	fasync_aux(-1, file, 0);
	if (--aux_count) {
		unlock_kernel();
		return 0;
	}
	kbd_write_cmd(AUX_INTS_OFF);			    /* Disable controller ints */
	kbd_write_command_w(KBD_CCMD_MOUSE_DISABLE);
	aux_free_irq(AUX_DEV);
	unlock_kernel();
	return 0;
}

/*
 * Install interrupt handler.
 * Enable auxiliary device.
 */

static int open_aux(struct inode * inode, struct file * file)
  pre(module_state == MS_INIT)
  changes(aux_count, cg_handle_mouse_event)
{
	if (aux_count++) {
		return 0;
	}
	queue->head = queue->tail = 0;		/* Flush input queue */
	if (aux_request_irq(keyboard_interrupt, AUX_DEV)) {
		aux_count--;
		return -EBUSY;
	}
	kbd_write_command_w(KBD_CCMD_MOUSE_ENABLE);	/* Enable the
							   auxiliary port on
							   controller. */
	aux_write_ack(AUX_ENABLE_DEV); /* Enable aux device */
	kbd_write_cmd(AUX_INTS_ON); /* Enable controller ints */
	
	mdelay(2);			/* Ensure we follow the kbc access delay rules.. */

	send_data(KBD_CMD_ENABLE);	/* try to workaround toshiba4030cdt problem */

	return 0;
}

/*
 * Put bytes from input queue to buffer.
 */

static ssize_t read_aux(struct file * file, char * buffer,
			size_t count, loff_t *ppos)
  pre(AND {
    module_state == MS_INIT;
    0 < file < objct && tag[file]==file_tag;
    // what to say about buffer?  TODO: I don't check anything about it...
  })
  changes(task_struct_state, inode_i_atime, mem, meta,
          aux_queue_tail)
{
	DECLARE_WAITQUEUE(wait, current);
	ssize_t i = count;
	unsigned char c;

	if (queue_empty()) {
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		//add_wait_queue(&queue->proc_list, &wait);   // TODO
repeat:
		invariant { 
		  global_inv;
                  0 < file < objct && tag[file]==file_tag;
		}
		set_current_state(TASK_INTERRUPTIBLE);
		if (queue_empty() && !signal_pending(current)) {
			schedule();
			goto repeat;
		}
                current->state = TASK_RUNNING;
	        //remove_wait_queue(&queue->proc_list, &wait);   // TODO
	}
	while (i > 0 && !queue_empty()) {
		invariant { 
		  global_inv;
                  0 < file < objct && tag[file]==file_tag;
		}
		c = get_from_queue();
		put_user(c, buffer++);
		i--;
	}
	if (count-i) {
		file->f_dentry->d_inode->i_atime = CURRENT_TIME;
		return count-i;
	}
	if (signal_pending(current))
		return -ERESTARTSYS;
	return 0;
}

/*
 * Write to the aux device.
 */

static ssize_t write_aux(struct file * file, const char * buffer,
			 size_t count, loff_t *ppos)
  pre(AND {
    module_state == MS_INIT;
    0 < file < objct && tag[file]==file_tag;
    // what to say about buffer?  TODO: I don't check anything about it...
  })
  changes(inode_i_mtime, cg_handle_mouse_event)
{
	ssize_t retval = 0;

	if (count) {
		ssize_t written = 0;

		if (count > 32)
			count = 32; /* Limit to 32 bytes. */
		do {
                        invariant { 
                          global_inv;
                          0 < file < objct && tag[file]==file_tag;
                        }
			char c;
			get_user(c, buffer++);
			aux_write_dev(c);
			written++;
		} while (--count);
		retval = -EIO;
		if (written) {
			retval = written;
			file->f_dentry->d_inode->i_mtime = CURRENT_TIME;
		}
	}

	return retval;
}

/* No kernel lock held - fine */
static unsigned int aux_poll(struct file *file, poll_table * wait)
  pre(module_state==MS_INIT)
{
	//poll_wait(file, &queue->proc_list, wait);    // TODO
	if (!queue_empty())
		return POLLIN | POLLRDNORM;
	return 0;
}

static int __init psaux_init(void)
  pre(module_state==MS_UNINIT)
  changes(cg_handle_mouse_event)
{
	int retval;

	if (!detect_auxiliary_port())
		return -EIO;

	if ((retval = misc_register(0 /*TODO: &psaux_mouse*/)))
		return retval;

	queue = (struct aux_queue *) kmalloc(sizeof(*queue), GFP_KERNEL);
	if (queue == NULL) {
		printk(KERN_ERR "psaux_init(): out of memory\n");
		misc_deregister(0 /*TODO: &psaux_mouse*/);
		return -ENOMEM;
	}
	memset(queue, 0, sizeof(*queue));
	queue->head = queue->tail = 0;
        //init_waitqueue_head(&queue->proc_list);  // TODO

        module_state == MS_INIT;

#ifdef INITIALIZE_MOUSE
	kbd_write_command_w(KBD_CCMD_MOUSE_ENABLE); /* Enable Aux. */
	aux_write_ack(AUX_SET_SAMPLE);
	aux_write_ack(100);			/* 100 samples/sec */
	aux_write_ack(AUX_SET_RES);
	aux_write_ack(3);			/* 8 counts per mm */
	aux_write_ack(AUX_SET_SCALE21);		/* 2:1 scaling */
#endif /* INITIALIZE_MOUSE */
	kbd_write_command(KBD_CCMD_MOUSE_DISABLE); /* Disable aux device. */
	kbd_write_cmd(AUX_INTS_OFF); /* Disable controller ints. */

	return 0;
}

#endif /* CONFIG_PSMOUSE */
