// sxlexer.cc            see license.txt for copyright and terms of use
// code for sxlexer.h

#include "sxlexer.h"       // this module


SXLexer::SXLexer(StringTable &s, char const *fname)
  : BaseLexer(s, fname)
{
  // prime this lexer with the first token
  tokenFunc(this);
}


SXLexer::~SXLexer()
{}


int SXLexer::tok(SXTokenType t)
{
  updLoc();
  sval = NULL_SVAL;     // catch mistaken uses of 'sval' for single-spelling tokens
  return t;
}


int SXLexer::svalTok(SXTokenType t)
{
  updLoc();
  sval = (SemanticValue)addString(yytext, yyleng);
  return t;
}


STATICDEF void SXLexer::tokenFunc(LexerInterface *lex)
{
  SXLexer *ths = static_cast<SXLexer*>(lex);

  // call into the flex lexer; this updates 'loc' and sets
  // 'sval' as appropriate
  ths->type = ths->yylex();
}


SXLexer::NextTokenFunc SXLexer::getTokenFunc() const
{
  return &SXLexer::tokenFunc;
}

string SXLexer::tokenDesc() const
{
  if (type == TOK_SYMBOL) {
    return stringc << tokenKindDesc(type) << ": " << (StringRef)sval;
  }
  else {
    return tokenKindDesc(type);
  }
}

string SXLexer::tokenKindDesc(int kind) const
{
  switch (kind) {
    default: xfailure("bad token code");
    case TOK_EOF:    return "TOK_EOF";
    case TOK_LPAREN: return "TOK_LPAREN";
    case TOK_RPAREN: return "TOK_RPAREN";
    case TOK_SYMBOL: return "TOK_SYMBOL";
  }
}


// EOF
