// term.h
// represent a term, i.e. a function symbol applied to some number of argument terms

#ifndef TERM_H
#define TERM_H

#include "objlist.h"     // ObjList
#include "str.h"         // string
#include "strtable.h"    // StringRef

class Term {
public:      // data
  // function or constant symbol or variable name
  StringRef name;

  // arguments
  ObjList<Term> args;

  // string table for the 'parse' functions; caller must set
  // to something non-NULL before using 'parse'
  static StringTable *parseStrTable;

public:      // funcs
  // create with some initial arguments; all are owner pointers
  explicit Term(StringRef name);
  Term(StringRef name, Term *a1);
  Term(StringRef name, Term *a1, Term *a2);
  Term(StringRef name, Term *a1, Term *a2, Term *a3);
  Term(StringRef name, Term *a1, Term *a2, Term *a3, Term *a4);
  Term(StringRef name, Term *a1, Term *a2, Term *a3, Term *a4, Term *a5);
  ~Term();
  
  // true if this term denotes a unification variable;
  // for now, unification variables are those that are capitalized
  bool isVar() const;

  // print as sexp
  string asString() const;
  void asString(stringBuilder &sb) const;

  // parse from sexp; return owner ptr
  static Term *parse(char const *sexp);

  // same, but return in 'endp' the pointer to the last character
  // that was parsed as this sexp, + 1
  static Term *parse(char const *sexp, char const *&endp);
  
  // clone, renaming unification variables
  //Term *clone(int id) const;
};


#endif // TERM_H
