// proof.cc
// code for proof.h

#include "proof.h"       // this module
#include "rule.h"        // Rule


// ----------------------- ProofNode ----------------------
ProofNode::ProofNode(EnumNode *n)
  : enumNode(n),
    child(n->rule->premises.count())
{
  child.setAll(NULL);
}


ProofNode::~ProofNode()
{
  for (int i=0; i < child.size(); i++) {
    if (child[i]) {
      delete child[i];
    }
  }
}


// ----------------------- ProofTree ----------------------
ProofTree::ProofTree(ProofNode *r)
  : subst(),       // empty substitution initially
    root(r)
{}

ProofTree::~ProofTree()
{
  delete root;
}


string ProofTree::getFormula(bool falseConcl) const
{
  stringBuilder sb;
  sb << "(IMPLIES (AND TRUE";

  getPremises(sb, root);

  sb << ") ";

  if (falseConcl) {
    // see if the premises imply false
    sb << "FALSE";
  }
  else {
    subst.itermSexp(sb, root->rule()->conclusion);
  }

  sb << ")";
  return sb;
}


void ProofTree::getPremises(stringBuilder &sb, ProofNode const *n) const
{
  ObjListIter<Term> iter(n->rule()->premises);
  for (int i=0; i < n->child.size(); i++, iter.adv()) {
    if (n->child[i]) {
      // has refined premise; recurse
      getPremises(sb, n->child[i]);
    }
    else {
      // unrefined premise; emit
      sb << " ";
      subst.itermSexp(sb, iter.data());
    }
  }
  xassert(iter.isDone());
}


// EOF
