// proof.h
// represent a proof tree

#ifndef PROOF_H
#define PROOF_H

#include "array.h"          // GrowArray
#include "unify.h"          // Unify
#include "enum.h"           // EnumNode

class Rule;                 // rule.h

// node in the proof tree
class ProofNode {
public:      // data
  // enumeration tree node that gave rise to this proof node
  EnumNode *enumNode;              // (serf)

  // array of subproofs; the array has as many entries as there
  // are premises in 'rule()'
  GrowArray<ProofNode*> child;     // (array of owner ptrs)

public:      // funcs
  ProofNode(EnumNode *n);
  ~ProofNode();

  // the rule used at this node
  Rule *rule() const { return enumNode->rule; }
};


// an entire proof tree; crucially, this includes the substitutions
// arising from unifying conclusions with premises
class ProofTree {
public:      // data
  // substitutions
  Unify subst;

  // root of the proof tree
  ProofNode *root;                 // (owner)

private:     // funcs
  void getPremises(stringBuilder &sb, ProofNode const *n) const;

public:      // funcs
  ProofTree(ProofNode *r);
  ~ProofTree();

  // get the formula "P => Q" where P is the conjunction of
  // unrefined premises and Q is the root's conclusion
  string getFormula(bool falseConcl = false) const;
};



#endif // PROOF_H
