// rule.h
// inference rule, i.e. an axiom of the theory

#ifndef RULE_H
#define RULE_H

#include "objlist.h"       // ObjList

class Term;                // term.h
  
// this rule is implicitly universally quantified over all the
// unification variables (capitalized names) that appear in it
class Rule {
public:
  // premises; they are atoms, but represented as terms anyway
  ObjList<Term> premises;

  // similarly for the conclusion
  Term *conclusion;        // (owner)

public:
  Rule(Term *concl);
  ~Rule();
             
  // make a deep copy, renaming all unification variables so they
  // end with the given index
  Rule *clone(int id) const;
};

      
// build rules by supplying conclusion, then hypotheses
Rule *makeRule0(Term *c);
Rule *makeRule1(Term *c, Term *h);
Rule *makeRule2(Term *c, Term *h1, Term *h2);


#endif // RULE_H
