// term.h
// represent a term, i.e. a function symbol applied to some number of argument terms

#ifndef TERM_H
#define TERM_H

#include "objlist.h"     // ObjList
#include "str.h"         // string

class Term {
public:      // data
  // function or constant symbol or variable name
  string name;

  // arguments
  ObjList<Term> args;

public:      // funcs
  // create with some initial arguments; all are owner pointers
  explicit Term(char const *name);
  Term(char const *name, Term *a1);
  Term(char const *name, Term *a1, Term *a2);
  Term(char const *name, Term *a1, Term *a2, Term *a3);
  Term(char const *name, Term *a1, Term *a2, Term *a3, Term *a4);
  Term(char const *name, Term *a1, Term *a2, Term *a3, Term *a4, Term *a5);
  ~Term();
  
  // true if this term denotes a unification variable
  bool isVar() const;

  // print as sexp
  string asString() const;
  void asString(stringBuilder &sb) const;

  // parse from sexp; return owner ptr
  static Term *parse(char const *sexp);

  // same, but return in 'endp' the pointer to the last character
  // that was parsed as this sexp, + 1
  static Term *parse(char const *sexp, char const *&endp);
  
  // clone, renaming unification variables
  Term *clone(int id) const;
};


#endif // TERM_H
