// unify.h
// term unification

#ifndef UNIFY_H
#define UNIFY_H

#include "ptrmap.h"          // PtrMap
#include "term.h"            // Term, StringRef, etc.


// encapsulate the algorithm, in particular the substitution state
class Unify {
public:      // data
  // current substitution; implemented with PtrMap because the
  // use of StringRefs for variable names means names map 1-1
  // with address of names
  PtrMap<char const, Term const> subst;

  // when set to true, we only do "matching", i.e. the right
  // argument ('y') is never considered to have unification
  // variables in it; this defaults to false
  bool matchMode;

private:     // funcs
  bool unifyVariable(StringRef var, Term const *t);
  bool occurs(StringRef var, Term const *t);
  void extendSubst(StringRef var, Term const *t);

public:      // funcs
  Unify();
  ~Unify();

  // main entry point
  bool unify(Term const *x, Term const *y);

  // secondary entry
  bool match(Term const *pattern, Term const *term)
    { matchMode = true; return unify(pattern, term); }

  // substitution queries
  bool isVar(Term const *t) const;
  bool bound(StringRef var) const;
  Term const *lookup(StringRef var) const;

  // emit under a substitution
  string termSexp(Term const *t) const;
  void itermSexp(stringBuilder &sb, Term const *t) const;

  // construct a new term, applying the substitution to the
  // given one (but otherwise yielding a clone)
  Term *substitute(Term const *t) const;

  // call this to clear the substitution, to unify new terms
  void reset();
};


bool unifies(Term const *x, Term const *y);


#endif // UNIFY_H
