// main.cc            see license.txt for copyright and terms of use
// entry-point module for a program that parses vclang input

#include <iostream.h>        // cout
#include <stdlib.h>          // exit, getenv
#include <stdio.h>           // printf

#include "trace.h"           // traceAddSys
#include "srcloc.h"          // SourceLocManager
#include "vclang.ast.gen.h"  // VCLang ast
#include "exc.h"             // xBase
#include "vc_env.h"          // VCEnv
#include "prover.h"          // Prover
#include "ckheap.h"          // malloc_stats
#include "strutil.h"         // plural
#include "nonport.h"         // getMilliseconds


int doit(int argc, char **argv)
{
  //traceAddSys("progress");
  //traceAddSys("parse-tree");

  char const *progname = argv[0];
  TRACE_ARGS();

  SourceLocManager mgr;

  VCEnv env;
  if (argc >= 3 && 0==strcmp(argv[1], "-b")) {
    env.blockOfInterest = argv[2];
    argc -= 2;
    argv += 2;
  }

  if (argc < 2) {
    cout << "usage: [-tr flags..] [-b block] " << progname << " filename.vc\n"
         << "  some useful flags:\n"
         << "    parse       print shift/reduce steps during parsing\n"
         << "    printTree   print the AST, then stop\n"
         << "  -b block: only prove things in the named block\n"
         ;
    return 0;
  }

  VCSequence *top = env.parseVCLang(argv[1]);

  if (tracingSys("printTree")) {
    // print the tree
    top->debugPrint(cout, 0);
    return 0;
  }

  // emit the VCs as Simplify-compatible s-expressions, and prove them
  long start = getMilliseconds();
  top->eval(env);
  long stop = getMilliseconds();

  delete top;

  if (env.failedProofs) {
    int n = env.failedProofs;
    cout << "there " << plural(n, "was") << " " << n
         << " failed " << plural(n, "proof") << "\n";
    return 2;
  }
  else {
    cout << "total proving time: " << (stop-start) << " ms\n";
    return 0;
  }
}


enum HeapWalkOpts heapWalker(void *block, int size)
{
  if (*((char*)block) != '\xBB') {
    printf("  %p: %d\n", block, size);
  }
  else {
    // it's been deallocated, but the walk still includes it
    // b/c DEBUG_HEAP actually disables free()
  }
  return HW_GO;
}


int main(int argc, char **argv)
{
  //malloc_stats();
  
  int ret;
  try {
    ret = doit(argc, argv);
  }
  catch (xBase &x) {
    cout << x << endl;
    return 4;
  }

  //malloc_stats();

  // look for leaks
  //printf("leaks:\n");
  //walkMallocHeap(heapWalker);

  return ret;
}
